package NTLSystem;

import com.sun.java.swing.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import NTLSystem.features.*;
import NTLSystem.support.*;

/**
 * An editable data dictionary is an abstract data dictionary which
 * contains editable data elements.  The dictionary has mechanisms for
 * to edit elements in the data dictionary, as long as those data
 * elements implement the EditableDataElement interface.
 * <P>
 * The data is stored in a StringSortedHashtable, which also
 * implements the ListModel interface.  This allows the
 * EditableDataDictionary elements to be put directly in a JList.
 * <P>
 * $Id: EditableDataDictionary.java,v 1.14 1998/10/16 01:56:00 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    Josh Sled [jsled@icsi.berkeley.edu]
 * @see       EditableDataElement
 * @version   $Revision: 1.14 $
 **/
public abstract class EditableDataDictionary
    implements Serializable, Cloneable
{
  // serialization stream unique identifier
  // recompute [with serialver] when this class changes.
  static final long serialVersionUID = -6686561251979318647L;

  /* ATTRIBUTES */

  /** The actual data stored. **/
  protected StringSortedHashtable dict;  

  /**
   * The list of elements in the dictionary [when editing].
   **/
  private transient JList l;
  /**
   * The choice of allowable types in the dictionary [when editing].
   **/
  private transient JComboBox c;
  /**
   * The editing frame [when editing].
   **/
  private transient JFrame f;



  /* CONSTRUCTORS */

  /**
   * Basic constructor for an EditableDataDictionary.
   **/
  public EditableDataDictionary()
  {
    dict = new StringSortedHashtable();
  }


  /* METHODS */

  /**
   * Gets an element from the dictionary.
   *
   * @param     key The key for this object.
   * @return    The EditableDataElement which corresponds to the given
   *            key.  Null if no such element exists.
   **/
  protected EditableDataElement get(String key)
  {
    return (EditableDataElement)dict.get(key);
  }

  /**
   * Sets a value in the dictionary.
   *
   * @param     key The key for this element.
   * @param     value The object which the key corresponds to.
   **/
  protected void set(String key, EditableDataElement value)
  {
      dict.put(key, value);
  }

  /**
   * Unsets a value in the dictionary.
   *
   * @param     The key for the value to unset.
   **/
  protected void unset(String key)
  {
    dict.remove(key);
  }

  /**
   * Test for key existance.  Returns true when the key exists in the
   * dictionary. 
   *
   * @param     key The key to test existance of.
   * @return    True if the key refers to an element in the
   *            dictionary, false otherwise.
   **/
  protected boolean containsKey(String key)
  {
    return dict.containsKey(key);
  }

  /**
   * Test for value existance.  Returns true when the value exists in
   * the dictionary.
   *
   * @param     value The value to test existance of.
   * @return    True if the value exists in the dictionary, false
   *            otherwise. 
   **/
  protected boolean contains(EditableDataElement value)
  {
    return dict.contains(value);
  }

  /**
   * Gets an enumeration of the keys of the elements in the data
   * dictionary.  For your convenience [and because we can], this
   * enumeration is in sorted order.
   *
   * @return    An Enumeration of they keys in the
   *            EditableDataDictionary [Strings].
   **/
  public Enumeration getElementKeys()
  {
    return dict.getSortedKeys();
  }


  /**
   * Allows for edit the data dictionary using the GUI.
   * <P>
   * Editing implies the ability to:
   *  <UL>
   *  <LI>Create a new element, edit it and add it to the dictionary.
   *  <LI>Edit existing elements.
   *  <LI>Remove existing elements.
   *  <LI>Clone existing elements and edit the clones.
   *  </UL>
   *
   * @param     wait True if the calling code wants the Edit method to
   *            return after the editing window is closed, false if
   *            the calling code wants the Edit method to return
   *            immediately.
   **/
  public void Edit(boolean wait)
  {
    f = new JFrame();
    JButton b;
    JPanel p;
    DictActionHandler dah = new DictActionHandler();

    GridBagLayout gbl = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();

    f.getContentPane().setLayout(gbl);

    gbc.fill = GridBagConstraints.HORIZONTAL;
    gbc.weightx = gbc.weighty = 1.0;
    gbc.gridwidth = GridBagConstraints.REMAINDER;

    // JList - The dictionary object is also a ListDataModel, which
    // allows the list to be dealt with much easier...
    // We want 20 lines and multiple selections allowed
    l = new JList( this.dict );
    // FIXME: Find a way to support double-click-list-element = edit.
    // l.addActionListener( dah );
    l.setVisibleRowCount(20);
    l.setSelectionMode( ListSelectionModel.MULTIPLE_INTERVAL_SELECTION );
    JScrollPane jsp = new JScrollPane(l);

    gbl.setConstraints(jsp, gbc);
    f.getContentPane().add(jsp);

    gbc.weightx = gbc.weighty = 0.0;

    // View/Edit/Remove buttons
    p = new JPanel();
    b = new JButton("View Selected");
    b.setActionCommand("VIEW_SEL");
    b.addActionListener(dah);
    p.add(b);

    b = new JButton("Edit Selected");
    b.setActionCommand("EDIT_SEL");
    b.addActionListener(dah);
    p.add(b);
    gbl.setConstraints(p, gbc);
    f.getContentPane().add(p);

    p = new JPanel();
    b = new JButton("Remove Selected");
    b.setActionCommand("REMOVE_SEL");
    b.addActionListener(dah);
    p.add(b);
    
    b = new JButton("Clone Selected");
    b.setActionCommand("CLONE_SEL");
    b.addActionListener(dah);
    p.add(b);

    gbl.setConstraints(p, gbc);
    f.getContentPane().add(p);

    // Create new choice/button
    p = new JPanel();
    c = new JComboBox();
    // Fill the choice with the possible types.
    Class possibleTypes[] = getAllowableTypes();
    for (int i=0; i<possibleTypes.length; i++) {
      // cname is a fully-qualified name of the class
      // [ie, foo.bar.baz.Name].  We only want the "Name" part of it.
      String cname = possibleTypes[i].getName();
      int lastDotIndex = cname.lastIndexOf('.');
      c.addItem( cname.substring(lastDotIndex+1, cname.length()) );
    } // end for over i
    p.add(c);

    b = new JButton("Create New Element");
    b.setActionCommand("CREATE_NEW");
    b.addActionListener(dah);
    p.add(b);

    gbl.setConstraints(p, gbc);
    f.getContentPane().add(p);

    // Create a "Dismiss" button at the bottom to close the window.
    p = new JPanel();
    b = new JButton("Dismiss");
    b.setActionCommand("DISMISS");
    b.addActionListener(dah);
    p.add(b);
    
    gbl.setConstraints(p, gbc);
    f.getContentPane().add(p);

    // Done creating the window.  Now pack'n'show.
    f.pack();
    f.setVisible(true);

    // Create a new WindowCloser to deal with that.
    // It will automagically add itself as a WindowListener.
    new WindowCloser(f, false);

    // If we're going to wait until the editing is done, do so.
    if (wait) {
      WindowWaiter ww = new WindowWaiter();
      f.addWindowListener(ww);
      ww.WaitUntilDone();
    } 
  }


  // Gets a set of Class object representing new Types which can
  // appear in the dictionary. 
  // This can be generated through a hard-coded return or reflection
  // [for FeatureDictionary: all objects which subclass
  // FeatureDescriptor]. 
  protected abstract Class[] getAllowableTypes();

  /**
   * Redefinition of Object.equals(Object)
   **/
  public boolean equals(Object o)
  {
    // test for instance and null
    if (! getClass().isInstance(o)) { return false; }
    EditableDataDictionary edd = (EditableDataDictionary)o;
    // test the hashtable equivalence
    if (! dict.equals(edd.dict)) { return false; }
    return true;
  }

  /**
   * Redefinition of Object.clone();
   **/
  public Object clone()
  {
      try {
	  EditableDataDictionary toRet =
	      (EditableDataDictionary)super.clone();
	  toRet.dict = (StringSortedHashtable)dict.clone();
	  return toRet;
      } catch (CloneNotSupportedException cnse) {
	  // This object is cloneable, as is the
	  // StringSortedHashtable.  Therefore, this shouldn't happen.
	  throw new InternalError( cnse.getMessage() );
      } // end try-catch
  }

  /**
   * Redefinition of Object.toString()
   **/
  public String toString()
  {
    StringBuffer toRet = new StringBuffer(200);
    toRet.append("EditableDataDictionary\n");
    toRet.append("item count = ");
    toRet.append(dict.size());
    toRet.append("\n");
    toRet.append(dict.toString());
    return toRet.toString();
  }

  // Externalizable
  public void readExternal(ObjectInput input) {}
  public void writeExternal(ObjectOutput output) {}


  /* INNER CLASSES [for editing] */

  class DictActionHandler implements ActionListener {

    public void actionPerformed(ActionEvent ae) {
      String acmd = ae.getActionCommand();
      if (acmd.equals("DISMISS")) {
	// Close the window.
	f.dispatchEvent(new WindowEvent(f,
					WindowEvent.WINDOW_CLOSING));
      } else if (acmd.equals("VIEW_SEL") ||
		 acmd.equals("EDIT_SEL") ||
		 acmd.equals("CLONE_SEL")) {
	// View or Edit the selected elements in the list.
	//   get item names
	Object selItems[] = l.getSelectedValues();
	//   view each on a seperate thread
	for (int i=0; i < selItems.length; i++) {
	  Thread t = null;
	  EditableDataElement ede = get( (String)selItems[i] );

	  if (acmd.equals("VIEW_SEL")) {
	    // Viewing the object
	    t = new ViewThread(ede);
	  } else if (acmd.equals("EDIT_SEL")) {
	    // Editing the object
	    t = new EditThread(ede);
	  } else if (acmd.equals("CLONE_SEL")) {
	    // Clone the object and edit it like it is new.
	    ede = (EditableDataElement)ede.clone();
	    t = new EditThread(ede, true);
	  }
	  t.start();
	} // end for over i.
      } else if (acmd.equals("REMOVE_SEL")) {
	// Remove the selected elements in the list.
	//   get item names
	Object selItems[] = l.getSelectedValues();
	//   remove items
	for (int i=0; i<selItems.length; i++) {
	  unset( (String)selItems[i] );
	} // end for over i
      } else if (acmd.equals("CREATE_NEW")) {
	// Create a new object for the dictionary from the
	// allowableTypes [current value of the choice]
	Class possTypes[] = getAllowableTypes();
	//   get combobox item name
	String toCreate = (String)c.getSelectedItem();
	// Define here to get rid of "may not be initalized" warning
	EditableDataElement ede = null;
	//   find corresponding Class
	boolean found = false;
	int index=0;
	for (; (index < possTypes.length) && (!found); index++) {
	  if (possTypes[index].getName().endsWith(toCreate)) {
	    found = true;
	  }
	}
	if (found) {
	  //   create new instance
	  try {
	    ede = (EditableDataElement)possTypes[index-1].newInstance();
	  } catch (InstantiationException ie) {
	    // Shouldn't happen...
	    throw new InternalError();
	  } catch (IllegalAccessException iae) {
	    // Shouldn't happen...
	    throw new InternalError();
	  } // end try-catch

	  // Do object editing on a seperate thread.
	  Thread t = new EditThread(ede, true);
	  t.start();
	} else {
	  new ErrorDialog(f,
			  "Cannot create object of type " + toCreate);
	}
      } else {
	// Double-click on list item.
	//   get list item name
	String name = acmd;
	//   find object in dictionary
	EditableDataElement ede = get(name);
	//   edit object
	Thread t = new EditThread(ede);
	t.start();
      } // end if-else if-...-else
    } // end actionPerformed(...)

  } // end inner class ActionHandler


  /**
   * Inner class to handle viewing/editing...
   **/
  class ViewThread
  extends Thread
  {
    JFrame viewF;
    EditableDataElement ede;

    public ViewThread(EditableDataElement ede) {
      super(ede.getKey() + " Viewing Thread");
      this.ede = ede;
    }

    public void run() {
      viewF = new JFrame(ede.getKey() + " Viewer");
      GridBagLayout gbl = new GridBagLayout();
      GridBagConstraints gbc = new GridBagConstraints();

      viewF.getContentPane().setLayout(gbl);

      gbc.fill = GridBagConstraints.BOTH;
      gbc.anchor = GridBagConstraints.NORTHWEST;
      gbc.weightx = gbc.weighty = 1.0;
      gbc.gridwidth = GridBagConstraints.REMAINDER;
      gbc.gridheight = GridBagConstraints.RELATIVE;
 
      Component edep = ede.getViewPanel();
      gbl.setConstraints(edep, gbc);
      viewF.getContentPane().add(edep);

      ActionListener al = new ActionListener() {
	public void actionPerformed(ActionEvent ae) {
	  String acmd = ae.getActionCommand();
	  if (acmd.equals("Dismiss")) {
	    viewF.dispatchEvent(new WindowEvent(viewF,
						WindowEvent.WINDOW_CLOSING));
	  }
	}
      }; // end ActionListener anonymous class

      JPanel p = new JPanel();
      JButton b = new JButton("Dismiss");
      b.addActionListener(al);

      p.add(b);

      gbc.weightx = gbc.weighty = 0.0;
      gbc.gridheight = GridBagConstraints.REMAINDER;

      gbl.setConstraints(p, gbc);
      viewF.getContentPane().add(p);

      new WindowCloser(viewF, false);

      viewF.pack();
      viewF.show();

    } // end run
  }; // end ViewThread


/**
 * Inner class to do EditableDataElement editing.
 **/
  class EditThread
  extends Thread
  implements ActionListener
  {
    EditableDataElement ede;
    JFrame editF;
    boolean newElem = false;
    String oldName;

    public EditThread(EditableDataElement ede)
    {
      super(ede.getKey() + " editing thread");
      this.ede = ede;
      oldName = ede.getKey();
    }

    /**
     * @param     ede The element to edit.
     * @param     newElem True if the element is new and needs to
     *            be added to the dictionary when done.  False if the
     *            element is already in the dictionary.
     **/
    public EditThread(EditableDataElement ede, boolean newElem)
    {
      this(ede);
      this.newElem = newElem;
    }

    public void run()
    {
      editF = new JFrame(ede.getKey() + " Editor");
      GridBagLayout gbl = new GridBagLayout();
      GridBagConstraints gbc = new GridBagConstraints();
      
      editF.getContentPane().setLayout(gbl);

      Component edep = ede.getEditPanel();
      gbc.anchor = GridBagConstraints.NORTHWEST;
      gbc.fill = GridBagConstraints.BOTH;
      gbc.weightx = gbc.weighty = 1.0;
      gbc.gridwidth = GridBagConstraints.REMAINDER;
      gbc.gridheight = GridBagConstraints.RELATIVE;

      gbl.setConstraints(edep, gbc);
      editF.getContentPane().add(edep);

      JPanel p = new JPanel();
      JButton b;

      b = new JButton("Ok");
      b.setActionCommand("OK");
      b.addActionListener(this);
      p.add(b);

      b = new JButton("Cancel");
      b.setActionCommand("CANCEL");
      b.addActionListener(this);
      p.add(b);

      gbc.fill = GridBagConstraints.HORIZONTAL;
      gbc.weightx = gbc.weighty = 0.0;
      gbl.setConstraints(p, gbc);
      editF.getContentPane().add(p);

      // This will automagically add itself as a window listener.
      new WindowCloser(editF, false);

      editF.pack();
      editF.show();
    }

    public void actionPerformed(ActionEvent ae) {
      String acmd = ae.getActionCommand();
      if (acmd.equals("OK")) {
	// Try to save the values, then close the window.
	try {
	  Vector cannotSaveReasons = new Vector(5);
	  ede.canSaveValues(cannotSaveReasons);
	  if (cannotSaveReasons.size() > 0) {
	    cannotSaveReasons.insertElementAt("Cannot save new element because...", 0);
	    String reasons[] = new String[cannotSaveReasons.size()];
	    cannotSaveReasons.copyInto(reasons);
	    JOptionPane.showMessageDialog(editF, reasons); 
	  } else {
	    ede.saveValues();
	    if ( ! newElem &&
		 (! ede.getKey().equals(oldName)) &&
		 (! oldName.equals("")) ) {
	      // Updating the list will be handled with a fired change
	      // event by unset. 
	      unset(oldName);
	      set(ede.getKey(), ede);
	    } // end if names unequal

	    if (newElem) {
	      // check to see if the element is not already in the
	      // dictionary.
	      if (containsKey(ede.getKey())) {
		String errorMsg = new String("An element with the name " +
					     ede.getKey() +
					     " already exists in the dictionary. " +
					     " Please edit the name and try again.");
		JOptionPane.showMessageDialog(f, errorMsg);
	      } else {
		// Add the data element...
		set(ede.getKey(), ede);
		// ... and close the window...
		editF.dispatchEvent(new WindowEvent(editF,
						    WindowEvent.WINDOW_CLOSING));
		// ... and release editing resources...
		ede.doneEditing();
	      } // end if-else
	    } else {
	      // Close the window...
	      editF.dispatchEvent(new WindowEvent(editF,
						  WindowEvent.WINDOW_CLOSING));
	      // ... and release editing resources...
	      ede.doneEditing();
	    } // end if-else ( new element )
	  } // end if-else
	} catch (NotBeingEditedException nbee) {
	  // won't happen, as we're currently editing.
	  throw new InternalError(nbee.getMessage());
	} catch (CannotSaveValuesException csve) {
	  // Shouldn't happen since we check above.
	  throw new InternalError(csve.getMessage());
	} // end try-catch
      } else if (acmd.equals("CANCEL")) {
	// Indicate that we were canceled.
	// Just close the window [without saving the values].
	editF.dispatchEvent(new WindowEvent(editF,
					    WindowEvent.WINDOW_CLOSING));
      }

    } // end actionPerformed

  } // end EditThread

}
