package NTLSystem.features;

import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.lang.*;
import java.util.Vector;
import NTLSystem.EditableDataElement;

/**
 * A feature-descriptor contains meta information that pertains
 * to all instantiations of a particular feature.
 * <P>
 * $Id: FeatureDescriptor.java,v 1.22 1998/10/16 01:57:05 jsled Exp $
   <BR>
 * Last editor: $Author: jsled $
 *
 * @author     Tajh Taylor [tajh@icsi.berkeley.edu]
 * @version    $Revision: 1.22 $
 **/

public abstract class FeatureDescriptor
implements Serializable, Cloneable, EditableDataElement
{
  /**
   * serialization Stream Unique Identifier.  Recompute [with
   * serialver] when this class changes.
   **/
  static final long serialVersionUID = -5150662162969198646L;

   // private instance variables

   /** The name for this feature descriptor; used as search key. */
   private String name;

   /** Whether or not this feature is probabilistic. */
   private boolean probabilistic;

  /**
   * The mutability of this object.  When objects are !mutable, they
   * can no longer be edited, and cannot be made mutable again.  All
   * editing requests should be handled as viewing requests.
   **/
  protected boolean mutable = true;

  /**
   * The defaultValue for the FeatureDescriptor.  If the FD is
   * probabilistic, this object will implement ProbDist.
   * If the FD is non-probabilistic, this object will be a descendent
   * of Number.
   **/
  //protected Object defaultValue;
  public Object defaultValue;

  /** 
   * The editing/viewing panel [when editing].
   **/
  protected transient FDViewer viewer;

   // public instance methods

   /** A no-arg constructor for Serialization/Externalization. */
   public FeatureDescriptor()
   {
      name = "";
      probabilistic = false;
      defaultValue = null;
   }

   /** A constructor for determinate feature descriptors.
       Since no default distribution is being provided, it assumes
       the feature is determinate.
       @param newName the new name for this feature descriptor. */
   public FeatureDescriptor(String newName)
   {
      name = newName;
      probabilistic = false;
      defaultValue = null;
   }

   // since a default distribution is provided, this constructor assumes
   // that the new feature is probabilistic
   /** A constructor for probabilistic feature descriptors.
       Since a default distribution is being provided, it assumes
       the feature is probabilistic.
       @param newName        the new name for this feature descriptor.
       @param newDefaultDist the new default distribution for this
                             feature descriptor. */
   public FeatureDescriptor(String newName, ProbDist newDefaultDist)
   {
      name = newName;
      probabilistic = true;
      defaultValue = newDefaultDist;
   }

   /** Implementation of a new clone() method.
       @return a newly constructed descriptor just like this one. */
   public Object clone()
   {
     try {
       FeatureDescriptor newFD = (FeatureDescriptor)super.clone();
       newFD.probabilistic = probabilistic;
       newFD.name = new String(name);
       // New feature descriptors are mutable until explictly set to
       // be immutable. 
       newFD.mutable = true;

       if (probabilistic) {
	 newFD.defaultValue = ((ProbDist)defaultValue).clone();
       } else {
	 // NOTE: This is ugly, but it should work forever.  There
	 // won't be too much variation in the types of objects
	 // contained in a feature descriptor.
	 if (defaultValue instanceof Double) {
	   newFD.defaultValue = Double.valueOf(defaultValue.toString());
	 } else if (defaultValue instanceof Integer) {
	   newFD.defaultValue = Integer.valueOf(defaultValue.toString());
	 } else if (defaultValue instanceof String) {
	   newFD.defaultValue = new String((String)defaultValue);
	 }
       }
       return newFD;
     } catch (CloneNotSupportedException cnse) {
       // Shouldn't happen as Cloneable is implemented above.
       throw new InternalError(cnse.getMessage());
     } catch (NumberFormatException nfe) {
       // Shouldn't happen as we're just creating new object of the
       // same type, but might...
       throw new InternalError(nfe.getMessage());
     } // end try-catch
     
   }

   // meta-info reporting methods

   /** Tell whether or not this descriptor is for a
       feature with ordered values.
       @return boolean flag indicating orderedness. */
   public abstract boolean isOrdered();

   /** Tell whether or not this descriptor is for a probabilistic
       feature.
       @return boolean flag indicating probabilisticity. */
   public boolean isProbabilistic()
   { 
      return probabilistic;
   }

   /** Set whether or not this descriptor is for a probabilistic
       feature.
       @param newProb boolean flag indicating new 
                      probabilisticity state. */
   protected void setProbabilistic(boolean newProb) 
   { 
      probabilistic = newProb;
   }

   /**
    * Get the default value for this FeatureDescriptor.  It is the
    * responsibility of the calling code to determine if this is a
    * value [Integer, Double, String, etc] or a ProbDist.
    *
    * @return    The default value for this descriptor.
    **/
  public Object getDefaultValue()
  {
    return defaultValue;
  }

   /** Get the name of this descriptor. 
       @return a string containing the name of this descriptor. */
   public String getName() 
   {
      return name; 
   }

  /**
   * Set the name of this descriptor.
   * <P>
   * NOTE: This should be private [or protected].  It needs to be
   * public for FDViewer.saveValues(), because inner-class inheritance
   * is screwed.  If it ever gets fixed, this can be made private
   * again.
   * <P>
   * <STRONG>Do not use this method.  It will be made private in the
   * future, and your code will break.</STRONG>
   *
   * @param     newName the new name for this descriptor.
   **/
   public void setName(String newName) 
   {
     name = new String(newName);
   }

   /** Test for equality with other descriptors.
       @param o any object to be tested.
       @return boolean flag indicating equality. */
   public boolean equals(Object o)
   {
      // instance/null checking
      if (! getClass().isInstance(o))
         return false;

      // else, check name, probabilistic, and probdist.
      FeatureDescriptor fd = (FeatureDescriptor)o;
      if (! name.equals(fd.getName()) )
         return false;
      if (probabilistic != fd.isProbabilistic()) 
         return false;
      if (probabilistic)
      {
	if (! defaultValue.equals(fd.defaultValue))
	   return false;
      }
      return true;
   }

   //===========================================================
   // Value access stuff
  
  /** Test whether the argument object contains a valid value for
       this particular feature.  Note that implementations should 
       report false for ProbDist object if the feature is determinate,
       and for definite value object if the feature is probabilistic.
       @param o any object to be tested.
       @return boolean flag indicating validity. */
   public abstract boolean isValidValue(Object o);

  /**
   * Converts the given String into the approrpriate value Object [if
   * possible].   This need not work for ProbDist values, only
   * deterministic values.
   *
   * @param     textVal A String representation of the Value.
   * @return    The appropriate object for this FeatureDescriptor.
   * @exception IllegalArgumentException if the string cannot be
   *            converted to the appropriate object.
   **/
  public abstract Object valueFromText(String textVal)
       throws IllegalArgumentException;

  /**
   * Converts the given value into a String.
   *
   * @param     val The value to convert into a String.
   * @return    A String version of the given value.
   **/
  public abstract String valueAsText(Object val);

  /** Return the value with the maximum likelihood,
      given a param.
      @param the distribution to draw from */
       //public abstract Object getMLValue(ProbDist distribution);

  /** Return a value at random according to the distribution.
      @param the distribution to draw from */
       //public abstract Object chooseValue(ProbDist distribution);
  
   // interface implementations

   /** Read in a descriptor written in human-readable text form.
       @param input An ObjectInput object from which the text will
                    be read. */
   public void  readExternal(ObjectInput input)
      throws IOException
   {
      // read the header
      //TODO readExternal() method needs a generic word parser class
   }

   /** Write the descriptor out in human-readable text form.
       @param output An ObjectOutput object to which the text will
                     be written. */
   public void  writeExternal(ObjectOutput output)
      throws IOException
   {
     // FIXME
      // write a header
      output.writeChars("FeatureDescriptor\n");

      // write the name
      output.writeChars("   name ");
      output.writeChars(name);
      output.writeChars("\n");

      // write the probabilisticity
      output.writeChars("   probabilistic ");
      output.writeBoolean(probabilistic);
      output.writeChars("\n");

      // write the distribution, if needed
      if (isProbabilistic())
      {
         output.writeChars("   ");
	 // FIXME.
         //defaultValue.writeExternal(output);
      }

      // write a footer
      output.writeChars("\n");
   }

  /**
   * Returns a displayable name for this FeatureDescriptor.
   *
   * @return    A displayable name for this FD.
   **/
  public abstract String getDisplayableName();

  /* EditableDataElement interface implementation */

  /**
   * Gets the key for this EditableDataElement.
   *
   * @return    A String key for this EditableDataElement.
   **/
   public String getKey()
   {
      return getName();
   }


  /**
   * Returns a Panel for viewing this FeatureDescriptor's data.
   * Concrete classes are responsible for returning a Panel containing
   * the viewing functionality for the default value or default
   * distribution.
   *
   * @return    A panel for viewing this FeatureDescriptor's data.
   **/
  public Component getViewPanel()
  {
    return getEditViewPanel(false);
  }

  /**
   * Returns a Panel for viewing this FeatureDescriptor's data.
   * Concrete classes are responsible for returning a Panel containing
   * the editing functionality for the default value or default
   * distribution.
   *
   * @return    A panel for editing this FeatureDescriptor's data.
   **/
  public Component getEditPanel()
  {
    return getEditViewPanel(true);
  }

  /**
   * Returns a Panel which can be used to either edit or view this
   * FeatureDescriptor's data.  The only difference is the status of
   * the editability of the text field, so both Edit and View are
   * basically the same method; therefore, this method.
   *
   * @param     edit True if this is an editing panel, false if a
   *            viewing panel.
   **/
  protected abstract JPanel getEditViewPanel(boolean edit);

  public void saveValues()
       throws NotBeingEditedException, CannotSaveValuesException
  {
    if (viewer == null) {
      throw new NotBeingEditedException("This object is not currently being edited.");
    }
    viewer.saveValues();
  }

  public void canSaveValues(Vector reasons)
       throws NotBeingEditedException
  {
    if (viewer == null) {
      throw new NotBeingEditedException("This object is not currently being edited.");
    }
    viewer.canSaveValues(reasons);
  }

  public void doneEditing()
  {
    viewer = null;
  }

}

/**
   * The FDViewer is a panel which contains the editing and viewing
   * functionality for FeatureDescriptors.  It is intended to be
   * flexible enough to support all subclasses of FeatureDescriptor.
   * <P>
   * The FDViewer is a JPanel which consists of a few editing
   * components [TextFields, Choices, etc], one optional panel which
   * subclasses can use for their specific editing options, and a
   * second panel, using a CardLayout, which contains two panels: one
   * for editing Probabilistic values and one for editing
   * Deterministic values.  These panels are arranged as such:
   * <PRE>
   * +----------------------------------------------------------------+
   * |  Name: |  JTextField                                      |    |
   * |  Type: |  JLabel                                          |    |
   * |  Mutability:    ( ) Mutable       ( ) Immutable                |
   * | +------------------------------------------------------------+ |
   * | |                                                            | |
   * | |                     Optional Panel                         | |
   * | |                                                            | |
   * | +------------------------------------------------------------+ |
   * |           ( ) Deterministic       ( ) Probabilistic            |
   * | +------------------------------------------------------------+ |
   * | |                                                            | |
   * | |       CardLayout-based panels [detPanel, probPanel]        | |
   * | |                                                            | |
   * | +------------------------------------------------------------+ |
   * +----------------------------------------------------------------+
   *
   * <P>
   * In addition, the class implements ActionListener, which allows it
   * to control the handling and manipulation of the Buttons, Choices
   * and other UI components [that use ActionListeners, of course].
   * <P>
   * <STRONG>                    NOTE:                      </STRONG>
   * Inner class inheritance is broken.  This leads to the workaround
   * of specifying the FeatureDescriptor being edited in the
   * constructor.  This class then keeps a reference to the
   * FeatureDescriptor in order to get access to certain methods and
   * to change the values when saving.  This workaround has caused
   * certain methods to be made public [instead of private, as they
   * should be].  These methods are noted as such.
   * <P>
   *
   * Author: Josh Sled [jsled@icsi.berkeley.edu]
   **/
abstract class FDViewer
    extends JPanel
    implements ActionListener
{

    /* ATTRIBUTES */
    
    /**
     * True if we're viewing, false if we're editing.
     **/
    boolean viewing = false;
    /**
     * The [GridBag]Layout and Constraints for this JPanel.
     **/
    GridBagLayout gbl;
    GridBagConstraints gbc;
    /**
     * The name TextField.
     **/
    JTextField nameTF;
    /**
     * The type label
     **/
    JLabel type;
    /**
     * The mutability radiobuttons
     **/
    JRadioButton mut;
    JRadioButton immut;
    /**
     * The [optional] middle Panel for subclass-specific data
     **/
    JPanel optPanel;
    /**
     * The deterministic/probabilistic radiobuttons.
     **/
    JRadioButton det;
    JRadioButton prob;
    /**
     * The Probabilistic editing panel.
     **/
    JPanel probPanel;
    /**
     * The Deterministic editing panel.
     **/
    JPanel detPanel;
    /**
     * The CardLayout-using Panel for editing prob/det values.
     **/
    JPanel probDetPanel;
    /**
     * The CardLayout they use.
     **/
    CardLayout probDetCL;
    /** 
     * The FeatureDescriptor which contains this inner class.
     * <P>
     * NOTE: This is a workaround because inner-class inheritance is
     * broken in the JDK.  To be more precise, it seems that
     * inner-class inheritance is underspecificed; subclasses of
     * FDViewer don't know their correct containing instance of
     * 'this'.  Therefore, we save it explicitly, and use our own
     * saved version instead of relying on 'this'.
     **/
    FeatureDescriptor fd;


    /**
     * Constructor for the FDViewer.  Just your basic constructor,
     * which goes through, calls all the abstract methods and sets up
     * the panels.
     *
     * @param     this_fd The FeatureDescriptor which we are editing.
     *            This is required because of the Inner-class
     *            inheritance problem mentioned above.
     * @param     edit Whether or not we're doing editing.  If false,
     *            only viewing is required.
     **/
    public FDViewer(FeatureDescriptor this_fd, boolean edit)
    {
	super();

	// Save the FeatureDescriptor.this workaround [see note on field
	// fd, above.
	fd = this_fd;

	// Set the convenience boolean 'viewing'.
	// We're in 'viewing' mode if we're not editing or not mutable.
	// DeMorgan's Law tells us:
	// (!editing) || (!mutable) ==> !(editing && mutable)
	viewing = !(edit && fd.mutable);

	// Create basic fields
	JLabel l;
	gbl = new GridBagLayout();
	gbc = new GridBagConstraints();
	probDetCL = new CardLayout();

	setLayout(gbl);

	// Create things
	createTopFields();
	createOptionalPanel();
	createProbDetPanels();

	// Create the prob/det radiobuttons in a ButtonGroup, since
	// they're mutually exclusive.
	ButtonGroup bg = new ButtonGroup();
	det = new JRadioButton("Deterministic");
	det.addActionListener(this);
	bg.add(det);
	prob = new JRadioButton("Probabilistic");
	prob.addActionListener(this);
	bg.add(prob);
	if (fd.isProbabilistic()) {
	    prob.setSelected(true);
	    det.setSelected(false);
	} else {
	    prob.setSelected(false);
	    det.setSelected(true);
	}

	// Disable the buttons if we're viewing
	if (viewing) {
	    det.setEnabled(false);
	    prob.setEnabled(false);
	}
	
	// Now that everything is created, add them to this Panel.
	gbc.anchor = GridBagConstraints.WEST;

	gbc.fill = GridBagConstraints.NONE;
	gbc.gridwidth = GridBagConstraints.RELATIVE;
	l = new JLabel("Name:");
	gbl.setConstraints(l, gbc);
	add(l);

	gbc.fill = GridBagConstraints.HORIZONTAL;
	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gbl.setConstraints(nameTF, gbc);
	add(nameTF);

	gbc.fill = GridBagConstraints.NONE;
	gbc.gridwidth = GridBagConstraints.RELATIVE;
	l = new JLabel("Type:");
	gbl.setConstraints(l, gbc);
	add(l);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gbl.setConstraints(type, gbc);
	add(type);

	gbc.gridwidth = GridBagConstraints.RELATIVE;
	l = new JLabel("Mutability:");
	gbl.setConstraints(l, gbc);
	add(l);


	JPanel MutImmutPanel = new JPanel();
	MutImmutPanel.add(mut);
	MutImmutPanel.add(immut);

	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gbl.setConstraints(MutImmutPanel, gbc);
	add(MutImmutPanel);

	// Add the optional panel if it exists and contains some
	// components.
	if ( (optPanel != null) &&
	     (optPanel.getComponentCount() > 0) ) {
	    gbl.setConstraints(optPanel, gbc);
	    add(optPanel);
	}

	// Det/Prob Radio Buttons
	// [ In their own center-aligned JPanel for
	//   nice-display-sake... ]
	JPanel detprob = new JPanel();
	detprob.add(det);
	detprob.add(prob);

	gbc.anchor = GridBagConstraints.CENTER;
	gbc.gridwidth = GridBagConstraints.REMAINDER;
	gbl.setConstraints(detprob, gbc);
	add(detprob);

	// Add the bottom CardLayout-based panel, if it exists.
	if (probDetPanel != null) {
	    gbl.setConstraints(probDetPanel, gbc);
	    add(probDetPanel);
	}
    }

    /**
     * This method creates the top-most fields in the FDViewer.
     **/
    public void createTopFields()
    {
	String initName;
	if ( (initName = fd.getName()).equals("") ) {
	    initName = new String("<unnamed>");
	}
	
	nameTF = new JTextField( initName, 25 );
	if (viewing) {
	    nameTF.setEditable(false);
	}

	type = new JLabel( getFDTypeName() );

	ButtonGroup mbg = new ButtonGroup();
	mut = new JRadioButton("Mutable");
	immut = new JRadioButton("Immutable");
	mbg.add(mut);
	mbg.add(immut);
	if ( fd.mutable ) {
	    mut.setSelected(true);
	    immut.setSelected(false);
	} else {
	    mut.setSelected(false);
	    immut.setSelected(true);
	}

	if ( viewing ) {
	    mut.setEnabled(false);
	    immut.setEnabled(false);
	}
      
    }

    /**
     * This method creates the Optional Panel in the middle of the
     * FDViewer.  It is not required that subclasses re-implement this
     * class, as it is an optional panel.  However, it is useful for
     * almost all subclasses.
     **/
    public void createOptionalPanel()
    {
	; // noop
    }


    /**
     * Creates the Probabilistic and Deterministic editing panels, and
     * the CardLayout-based panel which they are a part of.  This
     * method shouldn't need to be overridden by subclasses, rather
     * the <A HREF="#createProbPanel()">createProbPanel()</A> and
     * <A HREF="#createDetPanel()">createDetPanel()</A> methods should
     * probably be overridden.
     **/
    public void createProbDetPanels()
    {
	createProbPanel();
	createDetPanel();

	// Create a new CardLayout-using JPanel
	probDetPanel = new JPanel(probDetCL);

	// Add the two seperate panels, if they exist.
	if (probPanel != null) {
	    probDetPanel.add(probPanel, "Prob");
	}
	if (detPanel != null) {
	    probDetPanel.add(detPanel, "Det");
	}
	// Initially show the correct one.
	if (fd.isProbabilistic()) {
	    probDetCL.show(probDetPanel, "Prob");
	} else {
	    probDetCL.show(probDetPanel, "Det");
	}
    }

    /**
     * Creates the panel used for editing Probabilistic values for
     * this FeatureDescriptor.
     **/
    public abstract void createProbPanel();

    /**
     * Creates the panel used for editing Deterministic values for
     * this FeatureDescriptor.
     **/
    public abstract void createDetPanel();

    /**
     * Returns a printable String of the type of FeatureDescriptor
     * [such as "Foo Feature Descriptor"].
     *
     * @return The name of the FeatureDescriptor's type.
     **/
    public abstract String getFDTypeName();

    /**
     * ActionListener implementation.  Can be overridden by subclasses
     * for their own ActionEvent handling.
     *
     * @param     ae The ActionEvent to process.
     **/
    public void actionPerformed(ActionEvent ae)
    {
	String acmd = ae.getActionCommand();
	if (acmd.equals("Probabilistic")) {
	    // Show the Probabilistic editing panel
	    probDetCL.show(probDetPanel, "Prob");
	} else if (acmd.equals("Deterministic")) {
	    // Show the Deterministic editing panel
	    probDetCL.show(probDetPanel, "Det");
	}
    }

    /**
     * This method is called by the FeatureDescriptor's
     * <TT>canSaveValues(...)</TT> method.
     *
     * @see EditableDataElement
     **/
    public void canSaveValues(Vector reasons)
    {
	if (nameTF.getText().equals("")) {
	    reasons.addElement("The name of this FeatureDescriptor cannot be empty.");
	}
    }

    /**
     * This method is called by the FeatureDescriptor's
     * <TT>saveValues()</TT> method.
     *
     * @see EditableDataElement
     **/
    public void saveValues()
	throws CannotSaveValuesException
    {
	// Get the values from the fields.
	if (viewing) {
	    return;
	}
	// Save the name
	fd.setName( nameTF.getText() );
	// Save the mutability selection.
	if (mut.isSelected()) {
	    fd.mutable = true;
	} else {
	    fd.mutable = false;
	}
	// Save the deterministic/probabilistic [radiobutton] choice.
	fd.setProbabilistic(prob.isSelected());
    }

} // end inner class FDViewer
