package NTLSystem.features;

import java.awt.*;
import java.io.*;

/**
 * An interface for probability distributions.
 * <P>
 * $Id: ProbDist.java,v 1.12 1998/10/16 01:57:14 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author Josh Sled [jsled@icsi.berkeley.edu]
 * @version $Revision: 1.12 $
 **/
public interface ProbDist
extends Serializable, Cloneable
{
  /**
   * The serialization Stream Unique Identifier [SUID].  Recompute
   * this [with serialver] when an incompatible change is made.
   **/
  static final long serialVersionUID = 4755134882032045535L;
  /**
   * Report on the continuousness of the distribution.
   * 
   * @return    True if the distribution is continuous, false
   *            otherwise.
   **/
  public boolean isContinuous();

  /**
   * Returns the probability at a specific value.
   *
   * @param     fValue The value of interest.
   * @return    The probability at that value.
   **/
  public double getProb(double fValue);

  /**
   * A random value according to the distribution.
   *
   * @return    A random value according to the distribution.
   **/
  public Object chooseValue();

  /**
   * The maximum likelihood value of the distribution.
   *
   * @return    The maximum likelihood value of the distribution.
   **/
  public Object getMLValue();

  /**
   * Probability distributions must implement a public
   * equality-checking method. 
   *
   * @param     o The Object to compare to.
   * @return    True if the ProbDists are equal, false otherwise.
   **/
  public boolean equals(Object o);

  /**
   * Normalize the distribution.
   **/
  public void normalize();

  /**
   * Implementation of the Cloneable interface.
   **/
  public Object clone();

  /**
   * Read in the distribution from some human-editable format.
   *
   * @param     input The ObjectInput object to use for input.
   **/
  public void readExternal(ObjectInput input);

  /**
   * Pretty-print the distribution in a format suitable for human
   * reading and editing.
   *
   * @param     output The ObjectOutput object to use for printing the
   *            distribution.
   **/
  public void writeExternal(ObjectOutput output);

  /**
   * Gets a displayable name of this ProbabilityDistribution.
   *
   * @return    A displayable name for this ProbabilityDistribution.
   **/
  public String getDisplayableName();

  /**
   * Returns a Component with the ability to edit or view the data in
   * the distribution, controlled by the parameter flag.
   *
   * @param     edit True if editing ability is requested, false if
   *            only viewing ability is required.
   * @return    A Component with the ability to do as stated.
   **/
  public Component getEditViewPanel(boolean edit);

  /**
   * Returns a Component with the ability to edit the data in the
   * distribution in place.  IE: the values are saved as they are
   * typed.
   *
   * @return    A Component with the ability to do as stated.
   **/
  public Component getEditInplacePanel();

  /**
   * Saves the values in the editing panel.
   *
   * @exception NotBeingEditedException If this object is not
   *            currently being edited.
   * @exception CannotSaveValuesException If the values which this
   *            object is editing cannot be saved for whatever
   *            reason.
   **/
  public void saveValues()
       throws NotBeingEditedException, CannotSaveValuesException;

  /**
   * Indicates that editing the ProbDist is over, and editing
   * resources can be released.
   **/
  public void doneEditing();

}
