package NTLSystem.features;

import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.text.*;
import java.util.Random;

/**
 * A Uniform distribution.  Uniform Distributions are immutable.
 * <P>
 * $Id: UniformDist.java,v 1.6 1998/10/16 01:57:15 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author Tajh Taylor [tajh@icsi.berkeley.edu]
 * @version $Revision: 1.6 $
 **/
public class UniformDist 
implements ProbDist
{
  /**
   * The serialization Stream Unique Identifier.  Recompute this [with
   * serialver] when an incompatible change is made.
   **/
  static final long serialVersionUID = -8054341331451645376L;

  /* ATTRIBUTES */

  /** The bounds of the probability mass area **/
  private double lowerBound, upperBound;

  /** A PRNG for generating random values.  This is lazy-created when
   * needed. 
   * <P>
   * NOTE: This should be a "weak reference" so it can be gc'd if
   * necessary... 
   **/ 
  private Random prng;

  /**
   * The lower bound JTextField [when editing].
   **/
  private transient JTextField lbTF;
  /**
   * The upper bound JTextField [when editing].
   **/
  private transient JTextField ubTF;

  /* CONSTRUCTORS */

  /**
   * Constructs a "default" UniformDist, with bounds of [0.0, 1.0].
   * <P>
   * No-arg constructor for Serialization/Externalization.
   **/
  public UniformDist()
  {
    super();
    setBounds(0.0, 1.0);
  }

  /**
   * Constructs a new UniformDist with the specified bounds.
   *
   * @param     newLowerBound The initial lower bound
   * @param     newUpperBound The initial upperBound
   **/
  public UniformDist(double newLowerBound, double newUpperBound)
  {
    super();
    setBounds(newLowerBound, newUpperBound);
  }

  /* Distribution-specific value access methods */

  /**
     Sets the bounds.  If the upper bound is below the lower bound,
     the upper bound becomes equal to the lower bound.
   * @param     newLowerBound The initial lower bound
   * @param     newUpperBound The initial upperBound
   **/
  protected void setBounds(double newLowerBound, double newUpperBound)
  { 
     lowerBound = newLowerBound;
     if (newUpperBound < newLowerBound)
        upperBound = lowerBound;
     else
        upperBound = newUpperBound;
  }

  /**
   * @return     The lower bound for the Uniform Distribution
   **/
  public double getLowerBound()
  {
     return lowerBound;
  }

  /**
   * @return     The upper bound for the Uniform Distribution
   **/
  public double getUpperBound()
  {
     return upperBound;
  }

  /* Implementation of the ProbDist interface */

  /**
   * Returns the height of the curve at a specific value of interest in the
   * UniformDist.
   * 
   * @param     fValue The value of interest.
   * @return    The probability at that value.
   **/
  public double getProb(double fValue)
  {
     double height;
     
     if ((upperBound - lowerBound) == 0.0)
     {
        if (fValue == upperBound)
	   return Double.POSITIVE_INFINITY;
	else
	   return 0.0;
     }
     else
     {
	if ((lowerBound <= fValue) && (fValue <= upperBound))
	   return (1/(upperBound - lowerBound));
	else
	   return 0.0;
     }
  }

  /**
   * Returns the maximum likelihood value of the distribution.
   * In the case of uniform, this is non-unique, so it returns
   * the lower bound.
   * @return    The maximum likelihood value of the distribution.
   **/
  public Object getMLValue()
  {
     Double wrapper;

     wrapper = new Double(lowerBound);
     return (Object)wrapper;
  }

  
  /**
   * Randomly chooses a value from the distribution.
   *
   * @return     A randomly-chosen value from the distribution.
   **/
  public Object chooseValue()
  {
     Double wrapper;

     // special case
     if ((upperBound - lowerBound) == 0.0) {
       return new Double(upperBound);
     }
     
     // Lazy-create the PRNG
     if (prng == null)
     {
	// Seed the PRNG
	prng = new Random(System.currentTimeMillis());
     }
     
     // Get a random uniformly-distributed value
     double randVal = prng.nextDouble();

     // Scale it to this Uniform dist., and wrap it in a Double
     wrapper = new Double(randVal * (upperBound - lowerBound) + lowerBound);
     return (Object)wrapper;
  }

  /**
   * Redefinition of Object.equals(Object)
   **/
  public boolean equals(Object o)
  {
     // Do instance and null check
     if (! getClass().isInstance(o)) { return false; }
     
     // Cast for ease of reference
     UniformDist ud = (UniformDist)o;
     return ( (ud.getLowerBound() == this.getLowerBound()) &&
	      (ud.getUpperBound() == this.getUpperBound()) );
  }

  /**
   * "Normalizes" a UniformDist, which does nothing.
   **/
  public void normalize()
  {
    ; // noop
  }

  /**
   * Reports on the distribution's continuity.
   *
   * @return    True, since UniformDists are continuous.
   **/
  public boolean isContinuous()
  {
     return true;
  }

  /**
   * Redefinition of Object.toString()
   **/
  public String toString()
  {
    StringBuffer sb = new StringBuffer(30);
    sb.append("UniformDist{");
    sb.append("lowerBound=");
    sb.append(getLowerBound());
    sb.append(", upperBound=");
    sb.append(getUpperBound());
    sb.append("}");
    return sb.toString();
  }

  /**
   * Redefinition of Object.clone()
   **/
  public Object clone()
  {
    try {
      UniformDist toRet = (UniformDist)super.clone();
      toRet.setBounds(this.getLowerBound(), this.getUpperBound());
      return toRet;
    } catch (CloneNotSupportedException cnse) {
      // Shouldn't happen because Cloneable is implemented.
      throw new InternalError();
    }
  }

  /**
   * Reads the UniformDist from an ObjectInput implementer.
   * <P>
   * The format is specified in
   * <A HREF="#writeExternal(ObjectOutput)">writeExternal(...)</A> below.
   *
   * @param     input The ObjectInput object to read from.
   **/
  public void readExternal(ObjectInput input)
  {
    // FIXME
  }
  
  /**
   * Writes a UniformDist in human-readable, human-editable format to
   * the given ObjectOutput.
   * <P>
   * Format is as follows:<BR>
   * FIXME
   *
   * @param     output The ObjectOutput object to write to.
   **/
  public void writeExternal(ObjectOutput output)
  {
    // FIXME
  }

  public String getDisplayableName()
  {
    return new String("Uniform Dist.");
  }

  /**
   * Gets an editing Component from the distribution which edits in
   * place.
   **/
  public Component getEditInplacePanel()
  {
    JPanel toRet = (JPanel)getEditViewPanel(true);
    KeyListener kl = new KeyAdapter() {
      public void keyTyped(KeyEvent ke) {
	try {
	  saveValues();
	} catch (NotBeingEditedException nbee) {
	  // Won't happen.
	} catch (CannotSaveValuesException csve) {
	  // What to do, what to do?
	  // FIXME
	}
      }
    };
    lbTF.addKeyListener(kl);
    ubTF.addKeyListener(kl);
    return toRet;
  }

  /**
   * Gets an edit or viewing panel from the distribution for
   * editing/viewing ability. 
   **/
  public Component getEditViewPanel(boolean edit)
  {
    JLabel l;
    GridBagLayout gbl = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();
    JPanel toRet = new JPanel(gbl);

    lbTF = new JTextField( (new Double(lowerBound)).toString(), 10 );
    ubTF = new JTextField( (new Double(upperBound)).toString(), 10 );

    if (! edit) {
      lbTF.setEditable(false);
      ubTF.setEditable(false);
    }

    gbc.fill = GridBagConstraints.NONE;
    gbc.gridwidth = GridBagConstraints.RELATIVE;
    l = new JLabel("Lower Bound:");
    gbl.setConstraints(l, gbc);
    toRet.add(l);

    gbc.fill = GridBagConstraints.HORIZONTAL;
    gbc.gridwidth = GridBagConstraints.REMAINDER;
    gbl.setConstraints(lbTF, gbc);
    toRet.add(lbTF);

    gbc.fill = GridBagConstraints.NONE;
    gbc.gridwidth = GridBagConstraints.RELATIVE;
    l = new JLabel("Upper Bound:");
    gbl.setConstraints(l, gbc);
    toRet.add(l);

    gbc.fill = GridBagConstraints.HORIZONTAL;
    gbc.gridwidth = GridBagConstraints.REMAINDER;
    gbl.setConstraints(ubTF, gbc);
    toRet.add(ubTF);

    return toRet;

  }

  public void saveValues()
       throws NotBeingEditedException, CannotSaveValuesException
  {
    if ( (lbTF == null) || (ubTF == null) ) {
      throw new NotBeingEditedException();
    }

    NumberFormat nf = NumberFormat.getNumberInstance();

    try {
      lowerBound = nf.parse(lbTF.getText()).doubleValue();
      upperBound = nf.parse(ubTF.getText()).doubleValue();
    } catch (ParseException pe) {
      throw new CannotSaveValuesException(pe.getMessage());
    } // end try-catch

  }

  public void doneEditing()
  {
    lbTF = ubTF = null;
  }

}
