package NTLSystem.support;

import java.io.*;
import java.util.*;

/**
 * NestedStreamTokenizer is like StreamTokenizer but adds a token type
 * which is a possibly-multi-line sub-string delimited by braces.
 * When such a token is found (TT_NESTED indicates this), the sval
 * field contains the nested string, sans braces.
 * This class also ignores Java-style // comments.
 * <P>
 * $Id: NestedStreamTokenizer.java,v 1.5 1998/07/31 21:07:00 jsled Exp $<BR>
 * Last editor:  $Author: jsled $
 *
 * @author    David Bailey [dbailey@icsi.berkeley.edu]
 * @author    Nancy Chang [nchang@icsi.berkeley.edu]
 * @see       StreamTokenizer
 * @version   $Revision: 1.5 $
 **/

public class NestedStreamTokenizer
    extends StreamTokenizer
{
  
  // {{{ ATTRIBUTES

  /** Constant for beginning of nested token **/
  private static final char OPEN_BRACE_CHAR = '{';

  /** Constant for ending of nested token **/
  private static final char CLOSE_BRACE_CHAR = '}';
  
  /** Constant for nested token **/
  public static final int TT_NESTED = -4;
  
  // }}} end attributes
    
  // {{{ CONSTRUCTORS
  public NestedStreamTokenizer(Reader r) {
    super(r);
    setUpSyntax();
  }
  
  // }}} end constructors


  // {{{ METHODS

    /**
     * Read next token.
     **/
  public int nextToken() throws IOException {
    int res = super.nextToken();
    if (res == OPEN_BRACE_CHAR) {
      // Hack: must disable parsing of numbers since re-converting
      // them to strings might introduce exponential notation which
      // Java cannot read back in...
      resetSyntax();
      StringBuffer sub = new StringBuffer();
      int nestDepth = 1;
      while (nestDepth > 0) {
	res = super.nextToken();
	if (res == OPEN_BRACE_CHAR) {
	  nestDepth++;
	}
	else if (res == CLOSE_BRACE_CHAR) {
	  nestDepth--;
	}
	else if (res == TT_EOF) {
	  throw new IOException("unexpected eof in nested stream");
	}
	if (nestDepth > 0) {
	  if (res == TT_WORD || res == TT_NUMBER) {
	    throw(new IOException("shouldn't be recognizing words or " +
				  "numbers in a nested block"));
	  }
	  else {
	    sub.append((char)res);
	  }
	}
      }
      sval = sub.toString();
      // Anti-Hack: Re-enable parsing of numbers, words, etc...
      setUpSyntax();
      res = TT_NESTED;
    }
    else if (res == TT_WORD) {
      // attempt to convert to number
      // (this is a workaround because Java can't read exponential notation)
      try {
	nval = Fmt.getDbl(sval);
	res = TT_NUMBER;
      }
      catch (NumberFormatException e) {
	// leave it as a TT_WORD
      }
    }
    return res;
  }

    /**
     * Specify syntax.
     **/
  private void setUpSyntax() {
    resetSyntax();
    slashSlashComments(true);
    //    parseNumbers();
    whitespaceChars((char)' ', (char)' ');
    whitespaceChars((char)'\t', (char)'\t');
    whitespaceChars((char)'\n', (char)'\n');
    eolIsSignificant(false);
    wordChars((char)'a', (char)'z');
    wordChars((char)'A', (char)'Z');
    wordChars((char)'0', (char)'9');
    wordChars((char)'-', (char)'-');
    wordChars((char)'_', (char)'_');
    wordChars((char)'.', (char)'.');  // for numbers...
  }
}
