package NTLSystem.support;

import java.io.*;
import java.util.*;
import NTLSystem.*;

/**
 * Priority Queue.  HIGHEST priorities returned first.
 * (Based on Lewis & Denenberg 1991, p. 303.)
 * <P>
 * $Id: PriorityQueue.java,v 1.2 1998/07/31 21:07:01 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    David Bailey [dbailey@icsi.berkeley.edu]
 * @author    Nancy Chang [nchang@icsi.berkeley.edu]
 * @version   $Revision: 1.2 $
 **/
public class PriorityQueue
    implements Serializable
{

    // {{{ ATTRIBUTES

    /** 
     * Maps Objects to Doubles
     **/
    private Hashtable priorities;

    /** 
     * Array of Objects arranged as a partially-ordered
     * tree by using [2i+1] and [2i+2] as children of [i]
     **/
    private Vector heap;

    // }}} end attributes
    
    // {{{ CONSTRUCTORS

    /**
     * Basic constructor for a PriorityQueue.
     **/
    public PriorityQueue() {
	priorities = new Hashtable();
	heap = new Vector();
    }

    // }}} end constructors

    // {{{ METHODS
    /**
     * Insert object with specified priority. 
     *
     * @param     o The Object to put in the PriorityQueue.
     * @param     p The priority of the Object being added.
     **/
    public void insert(Object o, double p) {
	priorities.put(o, new Double(p));
	heap.setSize(heap.size() + 1);
	int i = heap.size() - 1;
	while (i > 0 && 
	       ((Double)priorities.get(heap.elementAt((i-1)/2))).doubleValue() < p)
	    {
		heap.setElementAt(heap.elementAt((i-1)/2), i);
		i = (i-1)/2;
	    }
	heap.setElementAt(o, i);
    }

    /**
     * Insert object with specified priority. 
     *
     * @return    Object at 0 index.
     **/
    public Object peek() {
	if (heap.size() == 0) return null;
	return heap.elementAt(0);
    }

    /**
     * Remove Object from PriorityQueue.
     *
     * @return    Object removed from PriorityQueue.
     **/
    public Object remove() {
	if (heap.size() == 0) return null;
	Object res = heap.elementAt(0);
	priorities.remove(res);
	if (heap.size() == 1) {
	    heap.setSize(0);
	    return res;
	}
	double lastObjPri = 
	    ((Double)priorities.get(heap.elementAt(heap.size() - 1))).doubleValue();
	int i = 0;
	while (2*i+1 < heap.size() && 
	       ((Double)priorities.get(heap.elementAt(2*i+1))).doubleValue() >
	       lastObjPri 
	       ||
	       2*i+2 < heap.size() && 
	       ((Double)priorities.get(heap.elementAt(2*i+2))).doubleValue() >
	       lastObjPri) 
	    {
		int j;
		if (2*i+2 < heap.size()) {
		    if (((Double)priorities.get(heap.elementAt(2*i+1))).doubleValue() >
			((Double)priorities.get(heap.elementAt(2*i+2))).doubleValue()) {
			j = 2*i+1;
		    }
		    else {
			j = 2*i+2;
		    }
		}
		else {
		    j = heap.size() - 1;
		}
		heap.setElementAt(heap.elementAt(j), i);
		i = j;
	    }
	heap.setElementAt(heap.elementAt(heap.size() - 1), i);
	heap.setSize(heap.size() - 1);
	return res;
    }

    /**
     * Check whether PriorityQueue is empty.
     *
     * @return    True if PriorityQueue is empty.
     **/
    public boolean isEmpty() {
	return heap.size() == 0;
    }
}
