package NTLSystem.verblearn;

import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.*;
import java.io.*;
import java.util.*;
import NTLSystem.*;
import NTLSystem.features.*;
import NTLSystem.support.*;

/**
 * Params holds numerical parameters of the various algorithms, and the list
 * of feature virtual samples.
 * <P>
 * This code was taken and modified from dbailey's old VerbLearn.
 * <P>
 * <EM>This class needn't be public... it is only to get Javadoc 1.1
 * to generate docs for it.</EM>
 * <P>
 * $Id: Params.java,v 1.7 1998/10/16 01:59:37 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    David Bailey [dbailey@icsi.berkeley.edu]
 * @author    Nancy Chang [nchang@icsi.berkeley.edu]
 * @see       ScenarioSet
 * @version   $Revision: 1.7 $
 **/
public class Params
    implements PropertyCollection, Serializable
{

    // {{{ ATTRIBUTES

    /**
     * The serialization stream unique identifier.  Recompute this
     * [with serialver] if this class changes.
     **/
    static final long serialVersionUID = -3001444654780971889L;

    /** The name of this Params set. */
    private String name;
    
    /*
     * The actual parameters.
     */

    private Boolean adaptVirtuals;
    private static final String  adaptVirtualsDesc = "Whether to adapt Dirichlet prior in each slot during learning.";
    private Integer batchSize;
    private static final String  batchSizeDesc = "Number of new instances of a word to accumulate be4tween merging episodes.";
    private Integer maxTries;
    private static final String  maxTriesDesc = "None";
    private Double  maxVirtual;
    private static final String  maxVirtualDesc = "Prevent uniform distribution from causing infinitye virtuals.";
    private Double  minExplain;
    private static final String  minExplainDesc = "Minimum P(feature | sense) required to explain away feature."; 
    private Double  minLabel;
    private static final String  minLabelDesc = "Minimum P(word | action) required to emit word.";
    private Double  minMerge;
    private static final String  minMergeDesc = "Minimum similarity required between two senses if they are to be merged.";
    private Double  minObey;
    private static final String  minObeyDesc = "Minimum P(sense | worldstate) require to use sense in setting linking Fstruct.";
    private Double  minSetFeature;
    private static final String  minSetFeatureDesc = "Minimum \"peakedness\" of distribution required to set feature in linking Fstruct.";
    private Double  modelPriorWeight;
    private static final String  modelPriorWeightDesc = "Magnitude of preference for merging relative to preserving likelihood.";
    private Boolean testRecognition;
    private static final String  testRecognitionDesc = " (no description) ";
    private Integer trainingPasses;
    private static final String  trainingPassesDesc = "Replicate training data this many times.";
    private Boolean useWordFrequencies;
    private static final String  useWordFrequenciesDesc = " (no description) ";
    private Boolean verboseSenses;
    private static final String  verboseSensesDesc = " (no description) ";
    private Double  virtualInertia;
    private static final String  virtualInertiaDesc = "Effective number of samples supporting current virtuals when adapting them.";

    /**
     * An array of String parameter names and their associated types.
     * This is used by the getEditableProperties() method of the
     * PropertyCollection interface for reflection purposes.  The
     * Class objects, you'll note, are also the parameters of the 'set'
     * methods, and the setter method names can be derived from the
     * name Object.
     **/
    private static Object[] paramNames =
    { "adaptVirtuals", Boolean.class,
      "batchSize", Integer.class,
      "maxTries", Integer.class,
      "maxVirtual", Double.class,
      "minExplain", Double.class,
      "minLabel", Double.class,
      "minMerge", Double.class, 
      "minObey", Double.class,
      "minSetFeature", Double.class,
      "modelPriorWeight", Double.class,
      "testRecognition", Boolean.class,
      "trainingPasses", Integer.class,
      "useWordFrequencies", Boolean.class,
      "verboseSenses", Boolean.class,
      "virtualInertia", Double.class };

    /**
     * The virtual samples.  Maps String [Fstruct name.Feature name]
     * to Double [virtual sample count].
     **/
    private StringSortedHashtable virtuals;
    /**
     * A flag set to 'true' if the parameters have been modified.
     **/
    private boolean isModified = false;

    // }}} attributes

    // {{{ CONSTRUCTORS

    /**
     * Constructor for 'typical' Params.  Also, no-arg constructor for
     * serialization.  
     **/
    public Params()
    {
	adaptVirtuals = new Boolean( true );
	batchSize = new Integer( Integer.MAX_VALUE );
	maxTries = new Integer( 1 );
	maxVirtual = new Double( 10 );
	minExplain = new Double( 0.5 );
	minLabel = new Double( Math.pow(10, -7) );
	minMerge = new Double( 0 );
	minObey = new Double( Math.pow(10, -7) );
	minSetFeature = new Double( 1.5 );
	modelPriorWeight = new Double( 4.5 );
	testRecognition = new Boolean( false );
	trainingPasses = new Integer( 1 );
	useWordFrequencies = new Boolean( false );
	verboseSenses = new Boolean( false );
	virtualInertia = new Double( 50.0 );

	virtuals = new StringSortedHashtable();
    }
  

    /**
     * Creates a default Params object with the specified name.
     *
     * @param     name The name of this Params object.
     * @param     sset The ScenarioSet we're going to be interested in
     *            [for the Feature names and virtual samples values].
     **/
    public Params(String name, ScenarioSet sset)
    {
	this();
	setName( name );
	// Copy the virtual sample names.
	Vector templ = sset.getTemplateFstructs();
	Fstruct curFS;
	Vector fnames;
	String fstructName;
	String featureName;
	for (int i=0; i<templ.size(); i++) {
	    curFS = (Fstruct)templ.elementAt( i );
	    fstructName = curFS.getName(); // needed for name
	                                   // conversion
	    fnames = curFS.getAllFeatureNames();
	    for (int j=0; j<fnames.size(); j++) {
		featureName = (String)fnames.elementAt( j );
		virtuals.put( fstructName + "." + featureName,
			      new Double(1.0) );
	    } // end for over ea. Fstruct's Feature names
	} // end for over template Fstructs
    }

    /**
     * Constructor for Params from text input file. (temporary?)
     *
     * @param     input Input file, temporariliy "default" in params/.
     * @param     newId  Name of params set.
     **/
    public Params(NestedStreamTokenizer input,
		  String newId,
		  ScenarioSet sset,
		  FeatureDictionary fd)
    {
	name = newId;

	virtuals = new StringSortedHashtable();
	
	try {
	    int token = input.nextToken();
	    while (token != NestedStreamTokenizer.TT_EOF) {
		if (token != NestedStreamTokenizer.TT_WORD) {
		    throw(new IOException("expected parameter name"));
		}
		String p = input.sval;
		token = input.nextToken();
		if (p.equals("MinLabel") && 
		    token == NestedStreamTokenizer.TT_NUMBER) {
		    minLabel = new Double(input.nval);
		    setMinLabel( minLabel );
		} else if (p.equals("MinExplain") &&
			 token == NestedStreamTokenizer.TT_NUMBER) {
		    minExplain = new Double(input.nval);
		    setMinExplain( minExplain );
		} else if (p.equals("MinObey") && 
			   token == NestedStreamTokenizer.TT_NUMBER) {
		    minObey = new Double(input.nval);
		    setMinObey( minObey );
		} else if (p.equals("MinSetFeature") && 
			   token == NestedStreamTokenizer.TT_NUMBER) {
		    minSetFeature = new Double(input.nval);
		    setMinSetFeature( minSetFeature );
		} else if (p.equals("UseWordFrequencies") &&
			   token == NestedStreamTokenizer.TT_WORD &&
			   (input.sval.equals("true") || input.sval.equals("false"))) {
		    useWordFrequencies = new Boolean(input.sval);
		    setUseWordFrequencies( useWordFrequencies );
		} else if (p.equals("MinMerge") &&
			   token == NestedStreamTokenizer.TT_NUMBER) {
		    minMerge = new Double(input.nval);
		    setMinMerge( minMerge );
		} else if (p.equals("MaxTries") &&
			   token == NestedStreamTokenizer.TT_NUMBER) {
		    maxTries = new Integer((int)input.nval);
		    setMaxTries( maxTries );
		} else if (p.equals("ModelPriorWeight") &&
			   token == NestedStreamTokenizer.TT_NUMBER) {
		    modelPriorWeight = new Double(input.nval);
		    setModelPriorWeight( modelPriorWeight );
		} else if (p.equals("BatchSize") &&
			   token == NestedStreamTokenizer.TT_NUMBER) {
		    batchSize = new Integer((int)input.nval);
		    setBatchSize( batchSize );
		} else if (p.equals("TrainingPasses") &&
			   token == NestedStreamTokenizer.TT_NUMBER) {
		    trainingPasses = new Integer((int)input.nval);
		    setTrainingPasses( trainingPasses );
		} else if (p.equals("AdaptVirtuals") &&
			   token == NestedStreamTokenizer.TT_WORD &&
			   (input.sval.equals("true") || input.sval.equals("false"))) {
		    adaptVirtuals = new Boolean(input.sval);
		    setAdaptVirtuals( adaptVirtuals );
		} else if (p.equals("VirtualInertia") &&
			   token == NestedStreamTokenizer.TT_NUMBER) {
		    virtualInertia = new Double(input.nval);
		    setVirtualInertia( virtualInertia );
		} else if (p.equals("MaxVirtual") &&
			   token == NestedStreamTokenizer.TT_NUMBER) {
		    maxVirtual = new Double(input.nval);
		    setMaxVirtual( maxVirtual );
		} else if (p.equals("VerboseSenses") && 
			   token == NestedStreamTokenizer.TT_WORD && 
			   (input.sval.equals("true") || input.sval.equals("false"))) {
		    verboseSenses = new Boolean(input.sval);
		    setVerboseSenses( verboseSenses );
		} else if (p.equals("TestRecognition") && 
			   token == NestedStreamTokenizer.TT_WORD && 
			   (input.sval.equals("true") || input.sval.equals("false"))) {
		    testRecognition = new Boolean(input.sval);
		    setTestRecognition( testRecognition );
		} else if (p.equals("MotorFeatures") &&
			   token == NestedStreamTokenizer.TT_NESTED) {
		    readFeatures("Motor", input.sval, sset, fd);
		} else if (p.equals("WorldFeatures") &&
			   token == NestedStreamTokenizer.TT_NESTED) {
		    readFeatures("World", input.sval, sset, fd);
		} else {
		    throw(new IOException("unknown parameter specified: " + p));
		}

		token = input.nextToken();
	    }
	}
	catch (IOException e) {
	    System.err.println("Params load-from-file exception: " + 
			       e.getMessage());
	}
	// MAKE SURE WE'VE SET EVERYTHING THAT IS OBLIGATORY?:
  }
  
  // }}} Constructors
  
  // {{{ METHODS

    /**
     * Read in features -- now for virtuals only, ignore rest.
     *
     * @param     fstructName The name of the Fstruct.
     * @param     s The name of the Feature.
     **/
    private void readFeatures(String fstructName,
			      String s,
			      ScenarioSet sset,
			      FeatureDictionary fd) 
	throws IOException 
    {
	Fstruct fs = new Fstruct(fstructName, fd );
	Vector featsToAdd = new Vector( 10 );
	Feature f;

	try {
	    NestedStreamTokenizer input = new 
		NestedStreamTokenizer(new StringReader(s));
	    int token = input.nextToken();
	    while (token != NestedStreamTokenizer.TT_EOF) {
		if (token != NestedStreamTokenizer.TT_WORD) {
		    throw(new IOException("expected feature name (tok=" + token + ")"));
		}
		String fname = input.sval;
		System.out.println("Got feature name: " + fname);

		token = input.nextToken();
		if (token != NestedStreamTokenizer.TT_WORD) {
		    throw(new IOException("expected 'on' or 'off' for feature (tok=" 
					  + token + ")"));
		}
		// FIXME: If 'off', then 'ignore' that Feature? -- ask nchang

		//        boolean isOn = (input.sval.toLowerCase().equals("on"));
		//	isOnMap.put(fname, new Boolean(isOn));

		//	if (isOn) nameOnVec.addElement(fname);

		token = input.nextToken();
		if (token != NestedStreamTokenizer.TT_NUMBER) {
		    throw(new IOException("expected virtual-sample count (tok=" +
					  token +")"));
		}
		//	virtMap.put(fname, new Double(input.nval));

		// Create a feature of the same name to add to the Fstruct
		f = fd.getFeature( fname );
		if (f == null) {
		    System.out.println("Feature returned for fname " +
				       fname + " is null.");
		} else {
		    System.out.println("Adding feature " + fname +
				       ": " + f.toString() );
		    featsToAdd.addElement( f );
		}
 
		System.out.println("Adding virtual sample value " +
				   fstructName + "." + fname + ": " +
				   (new Double(input.nval)).doubleValue() );
		virtuals.put(fstructName + "." + fname, new Double(input.nval));

		// Eat up the feature value names.
		token = input.nextToken();
		if (token != NestedStreamTokenizer.TT_NESTED) {
		    throw(new IOException("expected nested list of feature values " + 
					  "(tok=" + token + ", sval=" + input.sval + 
					  ")"));
		}

		token = input.nextToken();
	    } // end while
	} catch (IOException e) {
	    System.err.println("Params exception loading features: " + 
			       e.getMessage());
	} // end try-catch

	try {
	    fs.putFeatures( featsToAdd.elements() );
	} catch (NTLException ntle) {
	    // We don't expect an exception
	    throw new InternalError( ntle.getMessage() );
	} // end try-catch

	sset.addTemplateFstruct( fs );

  }
    
    /**
     * Get name of parameter set.
     *
     * @return    The Name of this parameter set.
     **/
    public String getName() { return name; }

    /**
     * Set name of parameter set.
     *
     * @param     newName New name of parameter set.
     **/
    public void setName(String newName) { name = newName; }

    /**
     * Parameters have been modified.
     *
     * @return     True if parameters have been modified.
     **/
    public boolean isModified() { return isModified; }

    /**
     * Set isModified field.
     *
     * @param     modified Params have been modified.
     **/
    public void setModified(boolean modified) { 
	isModified = modified;
    }

    // {{{ ACCESS PARAMETERS
    public boolean adaptVirtuals()
    { return adaptVirtuals.booleanValue(); }
    public void    setAdaptVirtuals(Boolean b)
    { adaptVirtuals = b; }

    public int     batchSize()
    { return batchSize.intValue(); }
    public void    setBatchSize(Integer i)
    { batchSize = i; }

    public int     maxTries()
    { return maxTries.intValue(); }
    public void    setMaxTries(Integer i)
    { maxTries = i; }

    public double  maxVirtual()
    { return maxVirtual.doubleValue(); }
    public void    setMaxVirtual(Double d)
    { maxVirtual = d; }

    public double  minExplain()
    { return minExplain.doubleValue(); }
    public void    setMinExplain(Double d)
    { minExplain = d; }

    public double  minLabel()
    { return minLabel.doubleValue(); }
    public void    setMinLabel(Double d)
    { minLabel = d; }

    public double  minObey()
    { return minObey.doubleValue(); }
    public void    setMinObey(Double d)
    { minObey = d; }

    public double  minSetFeature()
    { return minSetFeature.doubleValue(); }
    public void    setMinSetFeature(Double d)
    { minSetFeature = d; }
    
    public double  minMerge()
    { return minMerge.doubleValue(); }
    public void    setMinMerge(Double d)
    { minMerge = d; }

    public double  modelPriorWeight()
    { return modelPriorWeight.doubleValue(); }
    public void    setModelPriorWeight(Double d)
    { modelPriorWeight = d; }

    public boolean testRecognition()
    { return testRecognition.booleanValue(); }
    public void    setTestRecognition(Boolean b)
    { testRecognition = b; }

    public int     trainingPasses()
    { return trainingPasses.intValue(); }
    public void    setTrainingPasses(Integer i)
    { trainingPasses = i; }
    
    public boolean useWordFrequencies()
    { return useWordFrequencies.booleanValue(); }
    public void    setUseWordFrequencies(Boolean b)
    { useWordFrequencies = b; }

    public boolean verboseSenses()
    { return verboseSenses.booleanValue(); }
    public void    setVerboseSenses(Boolean b)
    { verboseSenses = b; }

    public double  virtualInertia()
    { return virtualInertia.doubleValue(); }
    public void    setVirtualInertia(Double d)
    { virtualInertia = d; }

    // }}} ACCESS PARAMETERS

    /**
     * Returns the Feature names stored in the virtual sample table
     * The feature names returned are in sorted order, fwiw.
     *
     * @return    An Enumeration of the Feature names in the virtual
     *            sample table. 
     **/
    public Enumeration featureNames() { 
	return virtuals.getSortedKeys();
    }

    /**
     * Returns the number of Features in the virtual samples table. 
     *
     * @return    The number of Features in the virtual samples
     *            table. 
     **/
    public int numFeatures() {
	return virtuals.size();
    }

    /**
     * Returns an Enumeration of the names of the parameters.
     *
     * @return     An Enumeration of the parameter names.
     **/
    public static Enumeration paramsNames()
    {
	return new Enumeration() {
	    private int i = 0;
	    public boolean hasMoreElements() {
		return (i < paramNames.length);
	    }
	    public Object nextElement() {
		// post-increment; returning the parameter name at
		// 'i' [before the increment].
		Object toRet = paramNames[i];
		// Skip over the class field.
		i += 2;
		return toRet;
	    }
	}; // end anonymous Enumeration.
    }

    // temporary to read in old params files
    public double featureVirtuals(String fname) {
	double tempMotor = featureVirtuals("Motor", fname);
	if (!(tempMotor == -1))
	    return tempMotor;
	return featureVirtuals("World", fname);
    }

    // Returns -1 if fname is invalid. *** change references to include 
    // fstructName;; should really throw exception instead
    public double featureVirtuals(String fstructName, String fname) {
	String fullName = fstructName + "." + fname;
	if (virtuals.containsKey(fullName)) {
	    return ((Double)virtuals.get(fullName)).doubleValue();
	} else {
	    return -1 ;
	}
    }

    /**
     * Returns the virtual count for the given named Feature.
     *
     * @param     fName The name of the Feature of interest.
     * @return    The virtual sample count for the given Feature.
     * @exception NoSuchFeatureException If the given named Feature
     *            does not exist in the virtual sample table.
     **/
    public double getVirtualSamples(String fName)
	throws NoSuchFeatureException
    {
	Double virtSampCount = (Double)virtuals.get( fName );
	if (virtSampCount == null) {
	    throw new NoSuchFeatureException( "Feature " + fName +
					      " not found in the virtual sample table." );
	}
	return virtSampCount.doubleValue();
    }

    /**
     * Sets the virtual sample count for the given named Feature.
     *
     * @param     fName The name of the Feature of interest.
     * @param     vSamp The virtual sample count to set.
     **/
    public void setVirtualSamples(String fName, Double vSamp)
    {
	virtuals.put( fName, vSamp );
    }


    public String toString() {
	StringBuffer res = new StringBuffer();
	res.append("// Parameter set: ").append(name).append("\n");

	res.append("\n\n");
	for (int i=0; i<paramNames.length; i+=2 ) {
	    String f = (String)paramNames[i];
	    try {
		res.append("\n").append(f).append("\t\t").
		    append( ( getClass().getDeclaredField(f).
			      get( this ) ).toString() );
	    } catch (Exception e) {
		// Shouldn't happen, since:
		// The field exists [no NoSuchFieldException]
		// The field is accessible: it's private, but we're in
		//   the class [no IllegalAccessException].
		// The object is non-null [no NullPointerException].
		// This object is indeed a Params [no
		//   IllegalArgumentException].
		throw new InternalError( e.getMessage() );
	    } // end try-catch
	} // end for over i [paramNames]

	//     res.append("\n\nMotorFeatures {");
	//     fEnum = motorNames.elements();
	//     while (fEnum.hasMoreElements()) {
	//       String f = (String)fEnum.nextElement();
	//       res.append("\n\t").append(f).append("\t")
	// 	.append(((Boolean)motorIsOn.get(f)).booleanValue() ? "on" : "off")
	// 	.append("\t").append(motorVirtuals.get(f)).append("\t{");
	//       for (int i = 0; i < ((Vector)motorValues.get(f)).size(); i++) {
	// 	if (i > 0) res.append(" ");
	// 	res.append(((Vector)motorValues.get(f)).elementAt(i));
	//       }
	//       res.append("}");
	//     }

	//     res.append("\n}\nWorldFeatures {");
	//     fEnum = worldNames.elements();
	//     while (fEnum.hasMoreElements()) {
	//       String f = (String)fEnum.nextElement();
	//       res.append("\n\t").append(f).append("\t")
	// 	.append(((Boolean)worldIsOn.get(f)).booleanValue() ? "on" : "off")
	// 	.append("\t").append(worldVirtuals.get(f)).append("\t{");
	//       for (int i = 0; i < ((Vector)worldValues.get(f)).size(); i++) {
	// 	if (i > 0) res.append(" ");
	// 	res.append(((Vector)worldValues.get(f)).elementAt(i));
	//       }
	//       res.append("}");
	//     }

	res.append("\n}\n");
	return res.toString();
    }

    // {{{ PropertyCollection impl.

    public String getLabel() 
    { return getName(); }

    /**
     * This implementaion of the getEditableProperties method will
     * return editable property descriptors for both the parameters
     * themselves and the virtual sample values.  The virtual sample
     * values are listed entirely after the parameter values.
     **/
    public EditablePropertyDescriptor[] getEditableProperties()
    {
	// Divide by two because there are two entires for every
	// param.
	Vector epds = new Vector( paramNames.length/2 +
				  virtuals.size() );
	EditablePropertyDescriptor[] toRet;
	String paramName;
	String fixedParamName;
	String paramDesc;
	Method getter, setter;
	Method vsGetter, vsSetter;
	Class getterParams[] = {};
	Class setterParams[] = new Class[1];
	try {
	    Class vsGetterParams[] = { Class.forName("java.lang.String") };
	    Class vsSetterParams[] = { Class.forName("java.lang.String"),
				       Class.forName("java.lang.Double") };
	    vsGetter = getClass().getMethod( "getVirtualSamples",
					     vsGetterParams );
	    vsSetter = getClass().getMethod( "setVirtualSamples",
					     vsSetterParams );
	} catch (NoSuchMethodException nsme) {
	    // Shouldn't happen, as we have the correct signature and
	    // method names.
	    throw new InternalError( nsme.getMessage() );
	} catch (ClassNotFoundException cnfe) {
	    // Shouldn't happen, as we have the correct [and standard]
	    // class names.
	    throw new InternalError( cnfe.getMessage() );
	} // end try-catch

	// Add EditablePropertyDescriptors for the Parameters
	for ( int i=0 ;
	      i<paramNames.length ;
	      i+=2 ) {
	    paramName = (String)paramNames[i];
	    setterParams[0] = (Class)paramNames[ i+1 ];
	    try {
		getter = getClass().getMethod( paramName,
					       getterParams );
		// Get the first letter of the param name, upcase it
		// and add 'set' to the front...
		// "adaptVirtuals" => "setAdaptVirtuals" 
		fixedParamName =
		    paramName.substring(0, 1).toUpperCase() +
		    paramName.substring(1);
		setter = getClass().getMethod( "set" + fixedParamName,
					       setterParams );
		// Get the field named "<paramName>Desc", and add it
		// as extra data to the paramName [for 
		paramDesc =
		    (String)(getClass().getDeclaredField( paramName +
							  "Desc" )).get(this);
		epds.addElement(new EditablePropertyDescriptor(paramName +
							       "\n" + paramDesc,
							       (Class)paramNames[ i+1 ],
							       getter,
							       setter) );
	    } catch (NoSuchMethodException nsme) {
		// Shouldn't happen, since the method exists above.
		throw new InternalError( nsme.getMessage() );
	    } catch (NoSuchFieldException nsfe) {
		// Shouldn't happen, since we've defined "Desc"
		// fields for all the Params.
		throw new InternalError( nsfe.getMessage() );
	    } catch (IllegalArgumentException iarge) {
		// Shouldn't happen, since we're asking the correct
		// object [this] for one of our field values.
		throw new InternalError( iarge.getMessage() );
	    } catch (IllegalAccessException iacce) {
		// Shouldn't happen, since we have access to our
		// fields. 
		throw new InternalError( iacce.getMessage() );
	    } // end try-catch
	} // end for over i


	// Add EditablePropertyDescriptors for the virtual samples.
	Enumeration sortedKeys = virtuals.getSortedKeys();
	String vsampName;
	while ( sortedKeys.hasMoreElements() ) {
	    vsampName = (String)sortedKeys.nextElement();
	    epds.addElement( new EditablePropertyDescriptor(vsampName,
							    Double.class,
							    vsGetter,
							    vsSetter,
							    vsampName));
	} // end while over virtual sample names

	// Return all the EditablePropertyDescriptors in the expected
	// manner. 
	toRet = new EditablePropertyDescriptor[ epds.size() ];
	epds.copyInto( toRet );
	return toRet;
    }

    // }}} PropertyCollection impl.

}
