package NTLSystem.verblearn;

import com.sun.java.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import NTLSystem.*;
import NTLSystem.features.*;
import NTLSystem.support.*;

/**
 * A Scenario holds the various pieces of information involved in an 
 * event.  Mainly this is a Vector of Fstructs (that parameterize
 * the actions and entities involved in the scene).
 * <P>
 *
 * The other important data here is the collection of Labels, which
 * are indexed by Language.
 * <P>
 *
 * $Id: Scenario.java,v 1.11 1998/10/16 01:59:39 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    Nancy Chang [nchang@icsi.berkeley.edu]
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       Fstruct
 * @version   $Revision: 1.11 $
 **/
public class Scenario
    implements Cloneable, Serializable
{
    // {{{ ATTRIBUTES

    /**
     * Serialization stream unique identifier.  Recompute [with
     * serialver] when this class changes.
     **/
    static final long serialVersionUID = -6971761106702427384L;

    /** The name of this Scenario. */
    private String name;
    
    /** The collection of Fstructs. These should be in the same order
	as the template fstructs from the ScenarioSet. */
    private Vector fstructs;

    /** The labels for this Scenario (cross-linguistic). Maps String
	[language] to VerbComplex [label]. */
    private Hashtable labels;
    
    /**
     * Meta-data properties for this Scenario.
     **/
    private Properties properties;

    // }}} attributes

    // {{{ CONSTRUCTORS

    /**
     * Constructor for basic Scenario with given template.
     * @param     template Vector of Fstructs for given Scenario.
     * @exception NTLException If template does not contain Fstructs.
     **/
    public Scenario(Vector template)
	throws NTLException
    {
	name = new String("");
	labels = new Hashtable(); // hashed on language
	properties = new Properties();
	fstructs = new Vector(template.size()); 
	
	Object retrieved;

	Enumeration e = template.elements();
	while (e.hasMoreElements())
	    {
		retrieved = e.nextElement();

		if (!(retrieved instanceof Fstruct))
		    throw new NTLException("Expected Fstructs in " +
					   "enumeration.");

		fstructs.addElement( ((Fstruct)retrieved).clone() );
	    }
    }
    
    /**
     * Constructs a Scenario with the given name and template.
     *
     * @param     name The name of this Scenario.
     * @param     template Vector of Fstructs for given Scenario.
     * @exception NTLException If thrown by <A
     *            HREF="#Scenario(Vector)">Scenario(Vector)</A>.
     **/
    public Scenario(String name, Vector template)
	throws NTLException
    {
	this(template);
	setName(name);
    }

    /**
     * Constructs a Scenario with the given name from text string.
     * TEMPORARY FIXME!!!
     *
     * @param     name The name of this Scenario.
     * @param     template Vector of Fstructs for given Scenario.
     * @exception NTLException If thrown by <A
     *            HREF="#Scenario(Vector)">Scenario(Vector)</A>.
     **/
    public Scenario(String name, Vector template, String s)
	throws NTLException
    {
      this(name, template);
      if (s == null || s.trim().equals("")) return;
      try {

	String featName = null;
	String fsName = null;

	NestedStreamTokenizer input = new 
	  NestedStreamTokenizer(new StringReader(s));
	int token = input.nextToken();
	// This is a vector of vectors with the Features names.  As we
	// add features from the String 's', we remove their names
	// from their Vector in this Vector.  At the end, we have a
	// Vectors which contain the names of Features which are _not_
	// supposed to be in the Fstructs, which we remove.
	Vector tfNames = new Vector( template.size() );
	for (int i=0; i<template.size(); i++) {
	    tfNames.addElement( ((Fstruct)fstructs.elementAt(i)).
				getAllFeatureNames()   );
	} // end for over i

	while (token != NestedStreamTokenizer.TT_EOF) {
	    if (token != NestedStreamTokenizer.TT_WORD) {
		throw(new IOException("expected feature name"));
	    }
	    featName = input.sval;

	  // figure out what kind of feature this is -- assume mutual exclusivity
	  // of feature names
	  Enumeration fsEnum = fstructs.elements();
	  int i = -1;
	  int correctIdx = 0;
	  while (fsEnum.hasMoreElements()) {
	    Fstruct currentFs = (Fstruct)fsEnum.nextElement();
	    i++;
	    if (currentFs.hasFeature(featName)) {
		correctIdx = i;
		fsName = currentFs.getName();
	    }
	  }
	  if (fsName == null) {
	    throw new IllegalArgumentException("Fstruct " + fsName +
					       " was not found in given ScenarioSet template");
	  }
	  
	  token = input.nextToken();
	  if (token != '=') {
	    throw(new IOException("expected '='"));
	  }
	  token = input.nextToken();
	   if (token != NestedStreamTokenizer.TT_WORD) {
	     throw(new IOException("expected feature value"));
	   }
	  
	   Feature f = getFstruct(fsName).getFeature(featName);
	   // Remove this Feature name from the list.
	   System.out.println( "Adding feature name " + featName +
			       ", correctIdx=" + correctIdx);
	   ((Vector)tfNames.elementAt(correctIdx)).removeElement( featName );
	   // We know that all Features in VerbLearn are Discrete.
	   DiscreteFeatureDescriptor dfd =
	       (DiscreteFeatureDescriptor)f.getDescriptor();
	   // We can do this because we know that all the Features in
	   // VerbLearn are DiscreteDists.
	   DiscreteDist dd = (DiscreteDist)f.getValue();
	   if (dd == null) {
	       System.out.println("Something is wrong: dd is null");
	       //dd = (DiscreteDist)((ProbDist)dfd.getDefaultValue()).clone();
	   }
	   // Get the index of the given value
	   int idx = dfd.getValueIndex( input.sval ).intValue();
	   if (idx == -1) {
	       throw new InternalError( "Unknown value name " +
					input.sval +
					" in Feature " +
					f.getName() + "." );
	   }
	   System.out.println("returned index " + idx +
			      " for value " + input.sval);
	   // Create a new distribution with all it's pmass on the
	   // selected item. 
	   double initVals[] = new double[ dd.getSize() ];
	   // Set the indicated value to have pmass of '1', all others
	   // to have pmass of '0'... this Feature then represents
	   // evidence of the given Feature value.
	   for (int k=0; k<initVals.length; k++) {
	       initVals[k] = (k == idx) ? 1.0 : 0.0;
	   } // end for over DiscreteDist
	   f.setValue( new DiscreteDist(initVals) );

	   token = input.nextToken();

	} // end while

	// Remove the Features we didn't insert.
	Vector fnamesToRemove;
	Fstruct curFS;
	String toRem;
	for (int i=0; i<template.size(); i++) {
	    fnamesToRemove = (Vector)tfNames.elementAt( i );
	    curFS = (Fstruct)fstructs.elementAt(i);
	    for (int j=0; j<fnamesToRemove.size(); j++) {
		toRem = (String)fnamesToRemove.
		    elementAt( j );
		System.out.println("Nulling feature value " + toRem +
				   ", i=" + i + "j=" + j);
		((Feature)curFS.getFeature( toRem )).setValue( null );
	    } // end for over j
	} // end for over i

      } catch (IOException e) {
	System.err.println("Fstruct-from-string exception1: " + e.getMessage());
      } /* catch (Exception e) {
	System.err.println("Fstruct-from-string exception2: " + e.getMessage());
	} */

      // temporary check...
      System.out.println(toString());
    }

    // }}} Constructors
  
    // {{{ METHODS

    // {{{ names

    /**
     * Sets the name of the Scenario to the given string.
     *
     * @param     newName The new name of the Scenario.
    **/
    public void setName(String newName)
    {
	name = newName;
    }

    /**
    * @return    The name of the Scenario.
    **/
    public String getName()
    {
	return new String(name);
    }

    // }}} names

    // {{{ properties -- copied from Fstruct. Do we need properties?

    /**
    * Gets a property from the Scenario.  The property is indexed by the
    * given name. If the property does not exist, null is returned.
    *
    * @param     propName The name of the requested property.
    * @return    The property's value, or null if the requested
    *            property does not exist.
    **/
    public String getProperty(String propName)
    {
	return (String)properties.get(propName);
    }

    /**
    * Sets a property in the Scenario.  The property is given by
    * propName, and set to the value propValue.
    *
    * @param     propName The name of the property to set.
    * @param     propValue The value to set the property to.
    **/
    public void setProperty(String propName, String propValue)
    {
	properties.put(propName, propValue);
    }

    /**
    * Un-sets the given property in the Scenario.
    * 
    * @param     prop The property name to unset.
    **/
    public void unsetProperty(String prop)
    {
	properties.remove(prop);
    }

    /**
    * Returns true if the given property is present in the Scenario,
    * false otherwise.
    *
    * @param     property The name of the property to check for
    *            existance.
    * @return    True if the property exists in the Scenario; false
    *            otherwise. 
    **/
    public boolean isProperty(String property)
    {
	return properties.containsKey(property);
    }

    /**
    * Returns an Enumeration of the property names in the properties
    * collection.  The values for these names can then be requested
    * from the <A HREF="#getProperty(String)">getProperty</A> method.
    *
    * @return    An Enumeration of the properties of this Scenario.
    **/
    public Enumeration getProperties()
    {
	return properties.keys();
    }

    // }}} properties

    // {{{ fstructs

    /**
     * Returns (sub)fstruct with specified Name.
     *
     * @return    The Fstruct from the Scenario.
     **/
    public Fstruct getFstruct(String fsName)
      throws IllegalArgumentException {
	Enumeration fsEnum = fstructs.elements();
	while (fsEnum.hasMoreElements()) {
	  Fstruct currentFs = (Fstruct)fsEnum.nextElement();
	  if ((currentFs.getName()).equals(fsName)) {
	    return currentFs;
	  }
	}
	// not there if get to this point
	throw new IllegalArgumentException("Fstruct " + fsName +
					   " was not found in given Scenario");
    }

    /**
     * Returns an array of the Fstructs in this Scenario.
     *
     * @return    An array of the Fstructs in this Scenario.
     **/
    public Fstruct[] getFstructs()
    {
	Fstruct toRet[] = new Fstruct[ fstructs.size() ];
	fstructs.copyInto( toRet );
	return toRet;
    }

    /**
     * Returns a 'unified' Fstruct, which is the union of all Fstructs
     * in the Scenario.   The Features in this unified Fstruct are
     * named 'Fstruct name.Feature name', and otherwise are clones of
     * the values in their respective Fstructs.
     * <P>
     * Note that the original Fstructs can be re-created by finding
     * the first index of '.' in each Feature name, and placing the 
     * Features into new appropriately-named Fstructs.
     *
     * @return    The union of the Fstructs in this Scenario.
     **/
    public Fstruct getUnifiedFstruct()
    {
	// Construct the Fstruct with the Scenario name and the
	// global FeatureDictionary.
	Vector featsToAdd = new Vector( 10 );
	Fstruct toRet = new Fstruct( getName(),
				     VerbLearn.getFeatureDict() );
	Fstruct curFS;
	Vector curFSFeats;
	Feature tmpFeat;
	for (int i=0; i < fstructs.size(); i++) {
	    curFS = (Fstruct)fstructs.elementAt( i );
	    try {
		curFSFeats =
		    curFS.getFeatures( curFS.getAllFeatureNames().elements() );
	    } catch (NTLException ntle) {
		// We expect neither NTLException or
		// NoSuchFeatureException, since we're asking for
		// Features that the Fstruct just told us it has.
		throw new InternalError( ntle.getMessage() );
	    }
	    // Make sure we have room to hold the Features.
	    featsToAdd.ensureCapacity( featsToAdd.size() +
				       curFSFeats.size() );
	    for (int j=0; j<curFSFeats.size(); j++) {
		// Collect the Features from this Fstruct, and modify
		// their names...
		tmpFeat =
		    (Feature)((Feature)curFSFeats.elementAt(j)).clone();
		tmpFeat.setName( curFS.getName() + "." +
				 tmpFeat.getName() );
		featsToAdd.addElement( tmpFeat );
	    } // end for over current Fstruct Features.
	} // end for over 'fstructs'

	// Now, put the Features into the Fstruct, and return it.
	try {
	    toRet.putFeatures( featsToAdd.elements() );
	} catch (NTLException ntle) {
	    // We don't expect either NTLException or
	    // NoSuchFeatureException...
	    throw new InternalError( ntle.getMessage() );
	} // end try-catch
	return toRet;
    }

    // }}} fstructs

    // {{{ labels

    /**
     * Sets the Scenario's label for specified language.  If the
     * label is null, the label will be unset [if it exists].
     *
     * @param     language The language to label in.
     * @param     label The label to set.
     **/
    public void setLabel(String language, VerbComplex label) {
	VerbComplex oldLabel = (VerbComplex)labels.get( language );
	if ( label == null ) {
	    labels.remove(language);
	} else {
	    labels.put(language, label);
	} // end if-else
    }

    /**
     * Returns Scenario's label for specified language.
     *
     * @param     language The language to check for label.
     * @return    This Scenario's label for specified language.
     **/
    public VerbComplex getLabel(String language)
    {
	return (VerbComplex)labels.get(language);
    }

    /**
     * Returns true if Scenario is labeled in specified language.
     *
     * @param     language The language to check for label.
     * @return    True if Scenario is labeld in specified language.
     **/
    public boolean isLabeled(String language) 
    {
	return (labels.get(language) != null);
    }
    // }}} labels
    
    // {{{ clone
    /**
     * Returns a field-by-field copy of the object.  If the component
     * objects are cloneable, their clones will be part of the new
     * object.
     *
     * @return     A clone of the current object.
     **/
    public synchronized Object clone()
    {
	// do the actual cloning;
	try {
	    Scenario toRet = (Scenario)super.clone();
	    toRet.name = getName();
	    toRet.fstructs = (Vector)fstructs.clone();
	    toRet.properties = (Properties)properties.clone();

	    return toRet;
	} catch (CloneNotSupportedException cnse) {
	    // Won't happen.  super.clone() will succeed...
	    throw new InternalError( cnse.getMessage() );
	}
    }
    // }}} clone

  // {{{ tostring

  /**
   * Redefinition of Object.toString().
   **/
  public String toString()
  {
    StringBuffer toRet = new StringBuffer(1000);
    toRet.append("Scenario {\n");
    toRet.append("  Name: "); toRet.append(getName()); toRet.append("\n");
    toRet.append("  Properties {\n");
    StringWriter sw = new StringWriter();
    PrintWriter pw = new PrintWriter(sw);
    properties.list(pw);
    toRet.append(sw.toString());
    toRet.append("  }\n");
    toRet.append("  Fstructs {\n");
    Enumeration e = fstructs.elements();
    for (; e.hasMoreElements(); ) {
      toRet.append( ((Fstruct)e.nextElement()).toString() );
      toRet.append("\n");
    }
    toRet.append(" }\n");

    toRet.append("  Labels {\n");
    e = labels.keys();
    for (; e.hasMoreElements(); ) {
      String lang = (String)e.nextElement();
      toRet.append("    " + lang + ": " + labels.get(lang));
      toRet.append("\n");
    }
    toRet.append("}\n");
    return toRet.toString();
  }

  // }}} tostring

    // }}} methods
}

