package NTLSystem.verblearn;

import com.sun.java.swing.*;
import com.sun.java.swing.border.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import NTLSystem.*;
import NTLSystem.features.*;
import NTLSystem.support.*;

/**
 * A ScenarioSet is a collection of Scenarios, with some other
 * associated data which is 'global' to the collection of Scenarios.
 * This data includes:
 * <UL>
 * <LI>The collection of languages for which labels exist.
 * <LI>A collection of
 * <A HREF="NTLSystem.verblearn.Lexicon.html">Lexicon</A>s which have
 * been generated based on the Scenarios in this ScenarioSet.
 * <LI>A unique collection of label choices for future labeling
 * <LI>A collection of 
 * <A HREF="NTLSystem.verblearn.Param.html">Param</A> objects which
 * apply to this ScenarioSet.
 * <LI>More... see the fields.
 * </UL>
 * <P>
 *
 * Perhaps most importantly, the ScenarioSet contains a Vector of
 * Fstructs which are the 'template' for all Scenarios in this
 * ScenarioSet.  These template Fstructs contain information [in their
 * properties] about which Features are currently 'ignored' in the
 * ScenarioSet.
 * <P>
 *
 * <STRONG>Note: Loading and saving routines do their own file handling, 
 * unlike other classes in the VerbLearn system.</STRONG>
 * <P>
 *
 * $Id: ScenarioSet.java,v 1.16 1998/10/16 01:59:40 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    Nancy Chang [nchang@icsi.berkeley.edu]
 * @see       AbstractListModel
 * @see       Scenario
 * @version   $Revision: 1.16 $
 **/
public class ScenarioSet
    extends AbstractListModel
    implements Serializable, Cloneable
{
    
    // {{{ ATTRIBUTES
    /** 
     * Serialization stream unique identifier recompute [with
     * serialver] when this class changes. 
     **/
    static final long serialVersionUID = -5126178739094602841L;

    /** The name of this ScenarioSet. **/
    protected String name;

    /** A description of this ScenarioSet. **/
    protected String description;
        
    /** The stored collection of Scenarios.  Probably the most
     * important thing in the ScenarioSet. Maps Strings [scenario
     * names] to Scenarios. **/
    protected StringSortedHashtable scenarios;
    
    /**
     * A template of Fstructs used in this ScenarioSet's Scenarios.
     * Vector<Fstruct>.
     **/
    protected Vector template;
    
    /**
     * The Languages this ScenarioSet is interested in.  By
     * convention, the Language in position 0 is the "default"
     * language.
     **/
    protected Vector langs;

    /**
     * The possible choices for labels in this ScenarioSet.  This
     * should be the set of all given Labels, indexed by Language.
     * Thus, this is a mapping of Strings [lang] to
     * StringSortedHashtable [choices].  These StringSortedHashtables
     * are mappings of Strings [labels] to Integers [use counts].
     **/
    protected StringSortedHashtable labelChoices;
    
    /**
     * A flag set to 'true' if the labels have been modified.
     **/
    protected boolean labelsModified;

    /**
     * The Lexicons which exist in this ScenarioSet.
     **/
    protected Vector lexicons;

    /**
     * The Params which has been defined in this ScenarioSet.
     * Maps String [Params name] to Params.
     **/
    protected StringSortedHashtable params;
    /**
     * The key of the current Params object.
     **/
    protected String currentParamsKey;

    /**
     * The Datasets which are defined in this ScenarioSet.
     * Maps Strings [Dataset name] to Dataset.
     **/
    protected StringSortedHashtable datasets;
    /**
     * The key of the currently-selected Dataset.
     **/
    protected String currentDatasetKey;

    // }}} END ATTRIBUTES
    
    // {{{ CONSTRUCTORS

    /**
     * Basic constructor for a ScenarioSet.
     **/
    public ScenarioSet()
    {
	scenarios = new StringSortedHashtable();
	template = new Vector();
	langs = new Vector();
	lexicons = new Vector();
	labelChoices = new StringSortedHashtable();

	params = new StringSortedHashtable( 2 );
	Params defParams = new Params();
	defParams.setName("default");
	params.put( defParams.getName(), defParams );
	currentParamsKey = defParams.getName();

	datasets = new StringSortedHashtable( 2 );
	currentDatasetKey = null;
    }
    
    /**
     * Constructor for a ScenarioSet based on a template (array of
     * Fstructs).
     *
     * @param     template A Vector of Fstructs which templatize this
     *            ScenarioSet.
     **/
    public ScenarioSet(Vector template)
    {
	this();
	this.template = template;
    }
    
    // }}} END CONSTRUCTORS

    // {{{ METHODS

    // {{{ name accessors

    /**
     * Get the value of name.
     *
     * @return    Value of name.
     **/
    public String getName() { return name; }
    
    /**
     * Set the value of name.
     *
     * @param     v Value to assign to name.
     **/
    public void setName(String v) { this.name = v; }

    // }}} name accessors

    // {{{ description accessors

    /**
     * Get the value of description.
     * @return    Value of description.
     **/
    public String getDescription() { return description; }
    
    /**
     * Set the value of description.
     * @param     v Value to assign to description.
     **/
    public void setDescription(String v) { this.description = v; }

    // }}} description accessors
    
    // {{{ adding/removing Scenarios
    
    /**
     * Adds a scenario with specified name to ScenarioSet. 
     * If this replaces an entry in the ScenarioSet, so be it.
     *
     * @param     sc The Scenario to put in the ScenarioSet.
     * @param     id The name of the Scenario to put in the ScenarioSet.
     **/
    public void addScenario(String id, Scenario sc) {
	scenarios.put(id, sc);
	// Update the label choices.
	Vector v = getLanguages();
	String lang;
	String label;
	for (int i=0; i < v.size(); i++) {
	    lang = (String)v.elementAt( i );
	    VerbComplex vc = sc.getLabel( lang );
	    if (vc != null) {
		label = vc.toString();
		if (label != null) {
		    addLabelChoice( label, lang );
		}
	    }
	} // end for over languages [i]
    }
  
    /**
     * Adds a scenario to the ScenarioSet.   If this replaces an
     * entry in the ScenarioSet, so be it.
     *
     * @param     sc The Scenario to put in the ScenarioSet.
     **/
    public void addScenario(Scenario sc) {
	addScenario(sc.getName(), sc);
    }

    /**
     * Removes a scenario from the ScenarioSet.  
     *
     * @param     id The name of the Scenario to remove from the
     *            ScenarioSet.
     * @param     sc The Scenario to remove from the ScenarioSet.
     **/
    public void removeScenario(String id, Scenario sc)
    {
	scenarios.remove( id );
	// Update the labelChoices.
	Vector langs = getLanguages();
	String lang;
	String label;
	for (int i=0; i<langs.size(); i++) {
	    lang = (String)langs.elementAt( i );
	    label = (String)sc.getLabel( lang ).toString();
	    if ( label != null ) {
		removeLabelChoice( label, lang );
	    } // end if
	} // end for over languages [i]
    }
  
    /**
     * Removes a scenario from the ScenarioSet.  
     *
     * @param     sc The Scenario to remove from the ScenarioSet.
     **/
    public void removeScenario(Scenario sc){
	removeScenario( sc.getName(), sc );
    }
    // }}} adding/removing scenarios

    // {{{ retrieving Scenarios/names

    /**
     * Test to see if a particular named Scenario is in the
     * ScenarioSet.
     *
     * @param     searchName The name of the Scenario to test
     *            for.
     * @return    True if the Scenario is in the
     *            ScenarioSet, false if not.
     **/
    public boolean containsScenario(String searchName)
    {
	return scenarios.containsKey(searchName);
    }

    /**
     * Test to see if the particular Scenario object is in the
     * ScenarioSet.
     *
     * @param     otherSc The Scenario to check for.
     * @return    true If the Scenario is in the
     *            ScenarioSet, false if not.
     **/
    public boolean containsScenario(Scenario otherSc)
    {
	return scenarios.contains(otherSc);
    }
    
    /**
     * Gets the Scenario with the given name from the
     * ScenarioSet. Returns null if the object does not exist in
     * the ScenarioSet.
     *
     * @param     searchName The name of the Scenario to search
     *            for.
     * @return    The Scenario with the given name, or null if
     *            it does not exist.
     **/
    public Scenario getScenario(String searchName)
    {
	return (Scenario)scenarios.get(searchName);
    }

    /**
     * Returns an Enumeration of all the Scenarios in the
     * ScenarioSet. 
     *
     * @return    An Enumeration of all the Scenarios.
     **/
    public Enumeration getAllScenarios()
    {
	return scenarios.elements();
    }
    
    /**
     * Returns an enumeration of names of Scenarios in ScenarioSet
     *
     * @return    An enumeration of names of Scenarios in the ScenarioSet. 
     **/
    public Enumeration names()
    {
	return scenarios.getSortedKeys();
    }

    /**
     * Returns a Vector of all the Scenarios names in the
     * ScenarioSet.  -- needed?
     *
     * @return    A Vector of the Scenario names in the ScenarioSet.
     **/
    // public Vector scenarioNames()
    // {
    // 	return scenarioNames;
    // }

    /**
     * Returns the count of the number of Scenarios in the
     * ScenarioSet
     *
     * @return    The number of Scenarios in the
     *            ScenarioSet. 
     **/
    public int countScenarios()
      {
	return scenarios.size();
      }
    // }}}} retreiving scenarios

    // {{{ language handling

    /**
     * Returns the languages associated with this ScenarioSet.
     *
     * @return    A Vector [of Strings] containing the languages
     *            supported by this ScenarioSet. 
     **/
    public Vector getLanguages()
    {
	return langs;
    }

    /**
     * Adds the named language to this ScenarioSet.  If it alreadys
     * exists, then this method silently noops.
     *
     * @param     lang The Language to add.
     **/
    protected void addLanguage(String lang)
    {
	if (! langs.contains(lang)) {
	    langs.addElement(lang);
	    labelChoices.put( lang, new StringSortedHashtable() );
	}
    }

    /**
     * Removes a language from the ScenarioSet.
     *
     * @param     lang The Language to remove.
     **/
    protected void removeLanguage(String lang)
    {
	langs.removeElement(lang);
	labelChoices.remove( lang );
	// FIXME: Remove all Lexicons in this language.
	// FIXME: Remove labels in this Lang from all scenarios.
    }

    // }}} language handling

    // {{{ lexicon handling
    
    /**
     * Gets the set of available Lexicon objects.
     *
     * @return    The ScenarioSet's Vector of Lexicons.
     **/
    public Vector getLexicons()
    {
	return lexicons;
    }

    /**
     * Gets the Lexicon with the given name.  If a Lexicon with the
     * given name does not exist, null is returned. 
     *
     * @param     lexiconName The name of the Lexicon of interest.
     * @return    The Lexicon with the given name in the ScenarioSet,
     *            or null if it does not exist.
     **/
    public Lexicon getLexicon( String lexiconName )
    {
	int idx = getLexiconIndex( lexiconName );
	Lexicon toRet = null;
	if ( idx != -1 ) {
	    toRet = (Lexicon)lexicons.elementAt(idx);
	}
	return toRet;
    }

    /**
     * Find the index of the given named Lexicon.  -1 is returned if
     * the named Lexicon doesn't exist.
     *
     * @param     lexName The name of the Lexicon of interest.
     * @return    The index of the given named Lexicon, or -1 if it
     *            doesn't exist.
     **/
    protected int getLexiconIndex(String lexName)
    {
	Lexicon curLex;
	for (int i=0; i<lexicons.size(); i++) {
	    curLex = (Lexicon)lexicons.elementAt(i);
	    if ( curLex.getName().equals(lexName) ) {
		return i;
	    }
	} // end for over lexicons
	return -1;
    }

    /**
     * Adds the given Lexicon to the ScenarioSet.  If a Lexicon with
     * the same name already exists, it is replaced.
     *
     * @param     lex The Lexicon to add.
     **/
    public void addLexicon(Lexicon lex)
    {
	int index = getLexiconIndex(lex.getName());
	if ( index == -1 ) {
	    lexicons.addElement( lex );
	} else {
	    lexicons.setElementAt( lex, index);
	}
    }

    /**
     * Removed the Lexicon with the given name from the ScenarioSet.
     * The removed Lexicon is returned.  If a Lexicon with the given
     * name doesn't exist in the ScenarioSet, this method returns
     * null. 
     *
     * @param     lexName The name of the Lexicon to remove.
     * @return    The removed Lexicon, or null if it doesn't exist.
     **/
    public Lexicon removeLexicon( String lexName )
    {
	int index = getLexiconIndex( lexName );
	Lexicon toRet = null;
	if ( index != -1 ) {
	    toRet = (Lexicon)lexicons.elementAt( index );
	    lexicons.removeElementAt( index );
	}
	return toRet;
    }

    // }}} lexicon handling

    // {{{ label handling

    /**
     * Returns labels used in ScenarioSet for specified language.
     *
     * @return    StringSortedHashtable of labels used in ScenarioSet 
     *             for specified language.
     **/
    public StringSortedHashtable getLabelChoices(String language)
    {
	return (StringSortedHashtable)labelChoices.get( language );
    }
    
    /**
     * Adds a lexeme to labelChoices list of ScenarioSet.  The
     * labelChoices list maps Strings [languages] to other
     * StringSortedHashtables which contain the actual label
     * choices. These StringSortedHashtables map Strings [labels] to
     * Integers, which are counts of the label's usage.  This is a
     * usage-count mechanism, and removed labels which have a new
     * usage count of 0 are removed from the labelChoices list.
     *  
     * @param    lex Label lexeme to add.
     * @param    language Language of lexeme.
     **/
    public void addLabelChoice(String lex, String language)
    {
	StringSortedHashtable langList =
	    (StringSortedHashtable)labelChoices.get( language );
	if ( langList == null ) {
	    langList = new StringSortedHashtable();
	    labelChoices.put( langList, language );
	}

	Integer i = (Integer)langList.get( lex );
	if (i == null) {
	    // If it doesn't exist, create a new count.
	    i = new Integer( 1 );
	} else {
	    // If it does exist, increment the count.
	    i = new Integer( i.intValue() + 1 );
	}
	langList.put( lex, i );
    }
    
    /**
     * Removes a lexeme from the labelChoices list of a ScenrioSet.
     * If it is detected that no other uses of that label are in
     * existance, the label choice is removed from the list.
     *
     * @param     lex Label lexeme to remove.
     * @param     language Language of lexeme.
     **/
    public void removeLabelChoice(String lex, String language)
    {
	StringSortedHashtable langList =
	    (StringSortedHashtable)labelChoices.get( language );
	if ( langList == null ) { return; }

	Integer i = (Integer)langList.get( lex );
	if ( i != null ) {
	    if ( i.intValue() > 1 ) {
		i = new Integer( i.intValue() - 1 );
		langList.put( lex, i );
	    } else {
		langList.remove( lex );
	    } // end if-else
	} // end if
    }
    
    /**
     * Set label of Scenario for specified language.
     * <strong>Note: assume for now label is parsed into verbcomplex.</strong>
     *
     * @param    sc Which scenario to label.
     * @param    label Label for given Scenario (currently verb complex)
     * @param    language Which language it's being labeled in.
     **/
    public void setLabel(Scenario sc,
			 String language,
			 VerbComplex label)
    {
	sc.setLabel(language, label);
    }


    // {{{ old: loadLabels()

    // FIXME from old! -- get to work FOR NOW. -- hardcoded paths!!
    public void loadLabels(String dirName, String language) {
      clearLabels(language);
      try {
	String fname = dirName + language + "1/labels";
	NestedStreamTokenizer input = 
	    new NestedStreamTokenizer( new BufferedReader(new FileReader(fname)));
	  int token = input.nextToken();
	  while (token != NestedStreamTokenizer.TT_EOF) {
	    if (token != NestedStreamTokenizer.TT_WORD) {
	      throw(new IOException("expected scenario name in labels"));
	    }
	    String scName = input.sval;
	    String scNum = scName.substring(2);
	    switch (scNum.length()) {
	    case 1:
		scNum = "00" + scNum;
		break;
	    case 2:
		scNum = "0" + scNum;
		break;
	    } // end switch
	    scName = "sc" + scNum;

	    token = input.nextToken();
	    if (token == NestedStreamTokenizer.TT_WORD) {
	      
	      if (scenarios.containsKey(scName)) {
		Scenario sc = (Scenario)scenarios.get(scName);
		// following allows "(none)" to indicate no label
		if (input.sval.equals("(none)")) {
		  sc.setLabel(language, null);
		} else {
		    VerbComplex vc = new VerbComplex(input.sval);
		  sc.setLabel(language, vc);
		  addLabelChoice( vc.toString(), language );
		}
	      }
	    } else {
	      throw new IOException("expected verb-complex in labels");
	    }
	    token = input.nextToken();
	  } // end while over labels
      } catch (IOException e) {
	System.err.println("ScenarioSet loadLabels exception: " + 
			   e.getMessage());
	clearLabels(language);
      }
    }

    // }}} old: loadLabels()

    // probably don't need?
    private void clearLabels(String language) {
	Enumeration scEnum = scenarios.elements();
	while (scEnum.hasMoreElements()) {
	    ((Scenario)scEnum.nextElement()).setLabel(language, null);
	}
    }
    
    // }}} label handling

    // {{{ template fstruct handling
    
    /**
     * Returns the Vector of template Fstructs.
     **/
    protected Vector getTemplateFstructs()
    {
	return template;
    }
    
    /**
     * Adds a new Fstruct to the Vector of template Fstructs.
     *
     * @param     f The new Fstruct to add to the template.
     **/
    protected void addTemplateFstruct(Fstruct f)
    {
	template.addElement(f);
    }

    /**
     * Removes an Fstruct from the Vector of template Fstructs.
     *
     * @param     f The Fstruct to remove.
     * @exception IllegalArgumentException If the Fstruct does not
     *            exist in the template list.
     **/
    protected void removeTemplateFstruct(Fstruct f)
	throws IllegalArgumentException
    {
	if ( ! template.removeElement(f) ) {
	    throw new IllegalArgumentException("Fstruct " +
					       f.getName() +
					       " was not found in the template for Scenario Set " + getName() );
	}
    }


     /**
     * Returns specified Fstruct; null if not there.
     * (should template be StringSortedHashtable?)
     *
     * @param     fstructName The Fstruct to retrieve.
     * @exception IllegalArgumentException If the Fstruct does not
     *            exist in the template list.
     **/
    public Fstruct getTemplateFstruct(String fstructName) 
      throws IllegalArgumentException
	{
	  Enumeration fsEnum = template.elements();
	  while (fsEnum.hasMoreElements()) {
	    Fstruct currentFs = (Fstruct)fsEnum.nextElement();
	    if (currentFs.getName().equals(fstructName)) {
	      return currentFs;
	    }
	  }
	  // if get to this point it's not there.
	  throw new IllegalArgumentException("Fstruct " +
					     fstructName +
					     " was not found in the template for Scenario Set " + getName() );
	}

     /**
     * Returns whether feature exists in given Fstruct (and is not "ignored").
     *
     * @param     f The Fstruct to remove.
     * @exception IllegalArgumentException If the Fstruct does not
     *            exist in the template list.
     **/
    public boolean featureExists(String fstructName, String fname) {
      try {
	Fstruct fs = getTemplateFstruct(fstructName);
	if (fs.hasFeature(fname) && (fs.getProperty(fname + ".ignored") != null))
	  return true;
      }
      catch(IllegalArgumentException e) {
      }
      return false;
    }
    // }}} template fstruct handling

    // {{{ from/to textfile

    /**
     * Read Scenarios in from a text file. -- FIXME taken from
     *  old Verblearn; needed ONLY TEMPORARILY (?), to read in old files!
     **/
    public void readScenarios(String specsfile) {
      NestedStreamTokenizer input;
      int token;
      try {
	// Create scenarios from specs:
	//input = DataSource.getNestedStreamTokenizer("scenario/specs");
	  try {
	      input = new NestedStreamTokenizer(new BufferedReader(new FileReader(specsfile)));
	  } catch (FileNotFoundException fnfe) {
	      System.out.println("Couldn't open file " + specsfile + ".");
	      return;
	  }
	token = input.nextToken();
	
	while (token != NestedStreamTokenizer.TT_EOF) {

	  // read in scenario name
	  if (token != NestedStreamTokenizer.TT_WORD) {
	    throw(new IOException("expected scenario name in specs"));
	  }
	  String scName = input.sval;
	  // Convert the name to '0'-padded three characters.
	  String scNum = scName.substring( 2 );
	  switch ( scNum.length() ) {
	  case 1:
	      scNum = "00" + scNum;
	      break;
	  case 2:
	      scNum = "0"+ scNum;
	      break;
	  } // end switch
	  scName = "sc" + scNum;
	  
	  // ignore initial fstruct; read other 
	  token = input.nextToken();
	  if (token == NestedStreamTokenizer.TT_NESTED) {
	    String initial = input.sval;
	    token = input.nextToken();
	    if (token != NestedStreamTokenizer.TT_NESTED) {
	      throw new IOException("expected final fstruct in fake scenario");
	    }
	    try {
	      Scenario scen = new Scenario(scName, template, input.sval);
	      //scenarios.put(scName, scen);
	      addScenario( scen );
	    } catch (NTLException e) {
	      System.err.println("Failed to build Scenario: " + e.getMessage());
	    }
	  } else {
	    throw new IOException("expected nested fstruct in specs");
	  }
	  token = input.nextToken();
	}
	
      } catch (IOException e) {
	System.err.println("ScenarioSet failed to load: " + e.getMessage());
      }
    }

    // }}} from/to textfile

    // {{{ toString
//     public String toString() {
// 	StringBuffer res = new StringBuffer().append("Scenario Set Summary:\n");
// 	Enumeration scnEnum = names();
// 	while (scnEnum.hasMoreElements()) {
// 	    Scenario sc = scenarios.get((String)scnEnum.nextElement());
// 	    res.append("\nScenario ").append(sc.getName()).append(":\n");
// 	    if (sc.hasLinks()) {
// 		res.append("initial = ").append(sc.initialLink());
// 		res.append("final = ").append(sc.finalLink());
// 	    }
// 	    if (sc.isLabelled()) {
// 		res.append("label = ").append(sc.label()).append("\n");
// 	    }
// 	}
// 	return res.toString();
//     }
    // }}} toString

    // {{{ incorporate
    /**
     * Incorporate a Vector of Scenarios (from the Generate panel).
     * If any of them have names equal to an existing scenario,
     * OVERWRITE the current scenario (including any labels).
     *  (warning below outdated now?)
     * BEWARE:  This may be undesired behavior, particularly if the
     * user has done a "compact names" operation in the Generate panel.
     *
     * @param     sv A vector of scenarios to add to this ScenarioSet.
     **/
    public void incorporate(Vector sv) {
	Enumeration sEnum = sv.elements();
	while (sEnum.hasMoreElements()) {
	  Scenario s = (Scenario)sEnum.nextElement();
	  scenarios.put(s.getName(), s);
	}
	// 	setChanged();
	// 	notifyObservers(); 
      }
    // }}} incorporate

    // {{{ merge
    /**
     * Merges another scenario set into this one. If it contains
     * a scenario with the same name, replace both scenario and labels
     * with those in new set.
     *
     * @param     ss ScenarioSet to merge with current ScenarioSet.
     **/
    // 
    public void merge(ScenarioSet ss){
      Enumeration sEnum = ss.names();
      while (sEnum.hasMoreElements()) {
	String name = (String)sEnum.nextElement();
	Scenario scen = ss.getScenario(name);

	if (scenarios.containsKey(name)) {
	  addScenario(scen);
	}
      }
    }

    // }}} merge

    // {{{ params

    /**
     * Returns the current Params object for the ScenarioSet.  This is
     * used by training to get it's tunable parameters.  This method
     * will return null if there is no currently-selected Params
     * object. 
     *
     * @return    The current Params object for the ScenarioSet, or
     *            null if there is no 'current' Params.
     **/
    public Params getParams()
    {
	Params toRet = null;
	if ( currentParamsKey != null ) {
	    toRet = getParams( currentParamsKey );
	}
	return toRet;
    }

    /**
     * Returns the Params object with the given name, or null if a
     * Params object with the given name does not exist.
     *
     * @param     pName The name of the Params object to return.
     * @return    The Params object with name pName.
     **/
    public Params getParams(String pName)
    {
	return (Params)params.get( pName );
    }

    /**
     * Add the given Params object to the ScenarioSet.  If a Params
     * object with the same name already exists in the ScenarioSet,
     * it's value is overwritten.
     *
     * @param     p The new Params object to add to the ScenarioSet.
     **/
    public void addParams( Params p )
    {
	params.put( p.getName(), p );
    }

    /**
     * Removes the Params object with the given name.  The object
     * removed is returned; null is returned if an object with the
     * given name did not exist in the ScenarioSet.
     *
     * @param     pName The name of the Params object to remove.
     * @return    The Params object removed from the ScenarioSet's
     *            collection of Params, or null if the named Params
     *            object does not exist.
     **/
    public Params removeParams( String name )
    {
	return (Params)params.remove( name );
    }

    // }}} params

    // {{{ dataset

    /**
     * Returns the current selected Dataset, which is used by
     * training. This method may return null if there is no
     * currently-selected Dataset .
     *
     * @return    The currently-selected Dataset, or null if no
     *            Dataset is 'current'.
     **/
    public Dataset getDataset()
    {
	Dataset toRet = null;
	if (currentDatasetKey != null ) {
	    toRet = getDataset( currentDatasetKey );
	}
	return toRet;
    }

    /**
     * Returns the Dataset with the given name, or null if no Dataset
     * with the given name exists.
     * 
     * @param     datasetName The name of the Dataset of interest.
     * @return    The Dataset with the given name.
     **/
    public Dataset getDataset( String datasetName )
    {
	return (Dataset)datasets.get( datasetName );
    }

    /**
     * Removes and returns the Dataset with the given name.  If no
     * Dataset with the given name exists, 'null' is returned.
     *
     * @param     dName The name of the Dataset to remove.
     * @return    The removed Dataset with the name 'dName', or null
     *            if no such Dataset exists.
     **/
    public Dataset removeDataset( String dName )
    {
	return (Dataset)datasets.remove( dName );
    }

    /**
     * Adds the given Dataset to the Dataset collection.  If a Dataset
     * with the same name exists in the collection, it is replaced
     * with the given Dataset.
     *
     * @param     dSet The Dataset to add to the Dataset collection.
     **/
    public void addDataset( Dataset d )
    {
	datasets.put( d.getName(), d );
    }

    /**
     * Returns an Enumeration of the Datasets in this ScenarioSet.
     *
     * @return    An Enumeration of the Datasets defined in this
     *            ScenarioSet.
     **/
    public Enumeration getDatasets()
    {
	return datasets.elements();
    }

    // }}} dataset

    // {{{ serialization
    
    /*
    private void readObject(java.io.ObjectInputStream in)
	throws IOException, ClassNotFoundException 
    {
	defaultReadObject( ois );
	// Go through the scenarios and add the labels to the
	// labelChoices StringSortedHashtable...

	// v is a Vector<StringSortedHashtable>
	Vector v = new Vector( langs.size() );
	for (int i=0; i<langs.size(); i++) {
	    
	} // end for
    }
    */

    // }}} serialization


    // {{{ AbstractListModel implementation
    public int getSize()
    {
	return scenarios.size();
    }

    public Object getElementAt(int i)
    {
	return scenarios.getElementAt(i);
    }
    // }}} AbstractListModel impl.

    // {{{ main
    /**
     * Testing code. -- fix! -nc
     **/
    public static void main(String args[])
    {
	// Get the properties.
	String propFN = new String("properties/verblearn-default.properties");
	Properties ourProps = new Properties(System.getProperties());
	try {
	    FileInputStream propsIn = new FileInputStream(propFN);
	    ourProps.load(propsIn);
	    System.setProperties(ourProps);
	} catch (SecurityException se) {
	    // Shouldn't happen as an app.
	    throw new InternalError(se.getMessage());
	} catch (FileNotFoundException fnfe) {
	    System.out.println("Could not find NTLSystem Verblearn properties file " + propFN);
	} catch (IOException ioe) {
	    System.out.println("Could not load default properties file.");
	} // end try-catch

	String scenSetName = ourProps.getProperty("ntlsys.verblearn.dataDir") +
	    "/" + ourProps.getProperty("ntlsys.verblearn.dictDir") +
	    "/" + ourProps.getProperty("ntlsys.verblearn.ssetDir") +
	    "/" + ourProps.getProperty("ntlsys.verblearn.ssetName");
	ScenarioSet ss = null;

	try {
	    FileInputStream fis =
		new FileInputStream(scenSetName);
	    ObjectInputStream ois = new ObjectInputStream(fis);
	    ss = (ScenarioSet)ois.readObject();
	} catch (IOException ioe) {
	    ioe.printStackTrace();
	    System.out.println(ioe.getMessage());
	} catch (ClassNotFoundException cnfe) {
	    // Won't happen
	}
    
	if (ss == null) {
	  System.out.println("Scenario Set _not_ restored.  " +
			     "Trying old specs file...");

	  scenSetName =
	      ourProps.getProperty("ntlsys.verblearn.dataDir") +
	      "/" + ourProps.getProperty("ntlsys.verblearn.scenDir") +
	      "/" + ourProps.getProperty("ntlsys.verblearn.scenSet");
	  System.out.println(": " + scenSetName + "\n");
	  ss = new ScenarioSet();
	  ss.readScenarios(scenSetName);
	}

// 	try {
// 	    FileOutputStream fos = new FileOutputStream(dictFN);
// 	    ObjectOutputStream oos = new ObjectOutputStream(fos);
// 	    oos.writeObject(ss);
// 	    oos.flush();
// 	    fos.close();
// 	} catch (IOException ioe) {
// 	    ioe.printStackTrace();
// 	    System.out.println(ioe.getMessage());
// 	}

	System.exit(0); 
    }
    // }}} main

    // }}} methods

    // {{{ Inner classes

    // }}} Inner classes
    
}
