package NTLSystem.verblearn;

import com.sun.java.swing.JTextArea;
import java.io.*;
import java.util.*;
import NTLSystem.*;
import NTLSystem.features.*;
import NTLSystem.support.*;

/**
 * Representation of a Word.  Mainly consists of a set of Senses.
 * <P>
 * <EM>This class needn't be public... it is only to get Javadoc 1.1
 * to generate docs for it.</EM>
 * <P>
 * $Id: Word.java,v 1.7 1998/10/16 01:59:45 jsled Exp $<BR>
 * Last editor: $Author: jsled $
 *
 * @author    David Bailey [dbailey@icsi.berkeley.edu]
 * @author    Joshua Sled [jsled@icsi.berkeley.edu]
 * @see       Sense
 * @version   $Revision: 1.7 $
 **/
public class Word
    implements Serializable
{

    /**
     * Convenience value: divisor for converting log to ln.
     **/
    public static final double LN_TO_LOG = Math.log(10);

    /**
     * The label of this Word.  The name is derived from this label by
     * the toString() method of VerbComplex.
     **/
    private VerbComplex label;

    /**
     * The collection of Senses which make up this Word.
     * Vector<Sense>.
     **/
    private Vector senses;
    
    /**
     * The candidate merges, given a new example.  Ordered by
     * similarity. 
     **/
    private PriorityQueue candidateMerges;

    /** Number of examples incorporated since last merge. **/
    private int backlog; 

    /**
     * A Vector of Fstructs seen over all Senses of this Word.
     * Used in computing likelihoods.
     **/
    private Vector examples;

    /**
     * Used to give unique names to senses.
     **/
    private int nextSenseNumber;

    /**
     * The Lexicon which contains this Word.
     **/
    private Lexicon lex;

    /**
     * A no-argument constructor for serialization.
     **/
    public Word()
    { }

    /**
     * Constructor for a Word.
     *
     * @param     label The label of this Word.
     * @param     l The Lexicon which contains this Word.
     **/
    public Word(VerbComplex label, Lexicon l)
    {
	this.label = label;
	senses = new Vector();
	candidateMerges = new PriorityQueue();
	examples = new Vector();
	nextSenseNumber = 1;
	lex = l;
    }

    /**
     * Get the name of this Word.  The name is a String representation
     * of the Word's label.
     *
     * @return    The name of this Word.
     **/
    public String getName() {return label.toString();}
    
    /**
     * Gets the label of this Word.  This is the label given by the
     * user during labeling.
     *
     * @return    The label of this Word.
     **/
    public VerbComplex getLabel() { return label; }

    /**
     * Set the label of this Word.
     *
     * @param     newLabel The new label for this Word.
     **/
    public void setLabel(VerbComplex newLabel) { label = newLabel; }

    /**
     * Returns the number of examples which this Word covers.
     **/
    public double numExamples() { return examples.size(); }
    
    /**
     * Returns the Lexicon which contains this Word, or null if this
     * Word is not contained in a Lexicon.
     *
     * @return    The Lexicon which contains this Word.
     **/
    public Lexicon getLexicon() { return lex; }

    /**
     * Sets the containing Lexicon for this Word.
     *
     * @param     lex The Lexicon which contains this Word.
     **/
    public void setLexicon(Lexicon lex) { this.lex = lex; }
	
    /**
     * Incorporate a new training instance, by adding a new sense for
     * it and possibly performing some merging (returns true iff so).
     * 
     * @param     f The Fstruct describing the action to incorporate.
     * @param     log The JTextArea to spit output to.
     **/
    public boolean incorporate(Fstruct f, JTextArea log) {
	examples.addElement(f);
	Sense s = new Sense(getName() + nextSenseNumber, f, this);
	log.append("Creating new sense " + s);
	nextSenseNumber++;
	backlog++;
	senses.addElement(s);
	Thread.currentThread().yield();
	addNewCandidateMerges(s);
	if ( backlog >=
	     getLexicon().getScenarioSet().getParams().batchSize() ) {
	    synch( log );
	    return true;
	} else {
	    return false;
	}
    }

    /**
     * Merge until we can't merge no more, then set the backlog to 0.
     *
     * @see       #mergeUntilLocalMax()
     * @param     log The JTextArea to use for output.  Not used if
     *            null.
     **/
    public void synch(JTextArea log) {
	mergeUntilLocalMax( log );
	backlog = 0;
    }

    /**
     * Find best merge candidates and perform the merge, repeating
     * until no satisfactory merge candidates are found or the
     * posterior drops.
     *
     * @param     log The JTextArea to output information to.  If
     *            null, it won't be used.
     **/
    private void mergeUntilLocalMax( JTextArea log ) {
	Thread.currentThread().yield();

	// Determine the posterior for the current Word model:

	double logPrior = logPrior();
	double logLike = logLikelihoodOfDataset();
	double oldLogPost = logPrior + logLike;
	double newLogPost = 0;  // this initialization is only to
                                // satisfy compiler
	double rate;
	Sense newSense = null;
	if ( getLexicon().getScenarioSet().getParams().testRecognition() ) {
	    rate  = recognitionRate();
	    String rateStr;
	    if (rate >= 0) {
		rateStr = String.valueOf((int)(rate * 100)) + "%";
	    } else {
		rateStr = "N/A";
	    }
	    if (log != null) {
		log.append("Starting merging for " + getName() + 
			   " (log prior=" + Fmt.dbl(logPrior) + 
			   ", log likelihood=" + Fmt.dbl(logLike) +
			   ", log posterior=" + Fmt.dbl(oldLogPost) + 
			   ", recog rate=" + rateStr +
			   "):\n");
	    }
	} else {
	    if (log != null) {
		log.append("Starting merging for " + getName() + 
			   " (log prior=" + Fmt.dbl(logPrior) + 
			   ", log likelihood=" + Fmt.dbl(logLike) +
			   ", log posterior=" + Fmt.dbl(oldLogPost) + 
			   "):\n");
	    }
	}

	// Repeatedly find and commit merges until no acceptable merge can
	// be found:
	while ( true ) {
	    Thread.currentThread().yield();

	    CandidateMerge cand;
	    boolean commitMerge = false;
	    int failedTries = 0;
	    Vector putBackList = new Vector();  // Vector<CandidateMerge>

	    // Try out merges until there are no remaining "plausible"
	    // (adequate similarity) merges on the queue, or MaxTries
	    // tentative merges have failed to raise the posterior:

	    while (true) {
		Thread.currentThread().yield();

		// Pick the highest-similarity candidate merge if such exists
		// and see if it's plausible enough:

		cand = (CandidateMerge)candidateMerges.remove();
		if (cand == null ||
		    cand.similarity < getLexicon().getScenarioSet().getParams().minMerge()) { 
		    if (senses.size() == 1) {
			if (log != null) {
			log.append("Stopped merging because only " +
				   "one sense remains.\n");
			}
		    } else {
			if (log != null) {
			log.append("Stopped merging because no (more) " +
				   "plausible merges were found.\n");
			}
		    }
		    break;
		}
		
		// Try out the candidate merge:

		newSense = merge(cand, log);
		logPrior = logPrior();
		logLike = logLikelihoodOfDataset();
		newLogPost = logPrior + logLike;
		if (getLexicon().getScenarioSet().getParams().testRecognition()) {
		    rate = recognitionRate();
		    String rateStr;
		    if (rate >= 0) {
			rateStr = String.valueOf((int)(rate * 100)) + "%";
		    } else {
			rateStr = "N/A";
		    }
		    if (log != null) {
			log.append("New " + getName() + 
				   " log prior=" + Fmt.dbl(logPrior) + 
				   ", log likelihood=" + Fmt.dbl(logLike) +
				   ", log posterior=" + Fmt.dbl(newLogPost)+
				   ", recog rate=" + rateStr +
				   ".\n");
		    }
		} else {
		    if (log != null) {
			log.append("New " + getName() + 
				   " log prior=" + Fmt.dbl(logPrior) + 
				   ", log likelihood=" + Fmt.dbl(logLike) +
				   ", log posterior=" + Fmt.dbl(newLogPost)+
				   ".\n");
		    }
		}

		// If an improvement, indicate the merge should be committed,
		// otherwise undo it and see if we should try again:

		if (newLogPost >= oldLogPost) {
		    commitMerge = true;
		    break;
		} else {
		    unmerge(cand, newSense, log);
		    putBackList.addElement(cand);
		    failedTries++;
		    if (failedTries ==
			getLexicon().getScenarioSet().getParams().maxTries()) {
			if (log != null) {
			    log.append("Stopped merging because the " + 
				       (getLexicon().getScenarioSet().getParams().
					maxTries() > 1 ?
					(String.valueOf(getLexicon().getScenarioSet().
							getParams().maxTries()) +
					 " most plausible merges ") : 
					"most plausible merge ") +
				       "reduced the posterior.\n");
			}
			break;
		    }
		}
	    }

	    // Restore bum merges since they might (?) work out in the future:
	    for (int i = 0; i < putBackList.size(); i++) {
		CandidateMerge cm = (CandidateMerge)putBackList.elementAt(i);
		candidateMerges.insert(cm, cm.similarity);
	    }

	    // Commit the Word model to the most-recent merge if indicated,
	    // else terminate the outer merging loop:

	    if (commitMerge) {
		purgeCandidateMerges(cand);
		addNewCandidateMerges(newSense);
		oldLogPost = newLogPost;
		if (failedTries > 0) {
		    if (log != null) {
			log.append("*** AN EXAMPLE OF KEEPING A " +
				   "NON-TOP-RANKED MERGE ***");
		    }
		}
	    } else {
		break;
	    }
	}
    }

    /**
     * Returns new sense to allow caller to undo merge with
     * unmerge(). 
     *
     * @param     cand The CandidateMerge to perform.
     * @param     log The JTextArea to use for output.  If null, this
     *            is not used.
     * @return    The new Sense resulting from the merge operation.
     **/
    private Sense merge(CandidateMerge cand, JTextArea log) {
	senses.removeElement(cand.s1);
	senses.removeElement(cand.s2);
	// Note: uses the current virtual samples for the new sense:
	Sense newSense = new Sense(cand.s1,
				   cand.s2,
				   getName() + nextSenseNumber,
				   this);
	nextSenseNumber++;
	senses.addElement(newSense);
	if (log != null) {
	    log.append("Merging " + cand.s1.getName() + " and " + 
		       cand.s2.getName() + " (similarity " + 
		       Fmt.dbl(cand.similarity) + ") to form " +
		       newSense);
	    }
	return newSense;
    }
    
    /**
     * Rolls back a merge operation.  Intended for use when a merge
     * lowers the posterior of the model. 
     *
     * @param     cand The CandidateMerge to undo.
     * @param     newSense The new Sense which should be removed, as
     *            it's creation lowered the posterior.
     * @param     log The JTextArea to use for output.  If null, this
     *            is not used. 
     **/
    private void unmerge(CandidateMerge cand,
			 Sense newSense,
			 JTextArea log)
    {
	senses.removeElement(newSense);
	senses.addElement(cand.s1);
	senses.addElement(cand.s2);
	if (log!= null) {
	    log.append("Posterior decreased; removing " +
		       newSense.getName() + " and restoring " + 
		       cand.s1.getName() + " and " +
		       cand.s2.getName() + ".\n");
	}
    }

    /**
     * Adds to candidateMerges all the new potential merges involving
     * a new Sense.
     *
     * @param     newSense The new Sense to add potential merges of.
     **/
    private void addNewCandidateMerges(Sense newSense) {
	Enumeration sEnum = senses.elements();
	while (sEnum.hasMoreElements()) {
	    Sense s = (Sense)sEnum.nextElement();
	    if (s != newSense) {
		CandidateMerge cand =
		    new CandidateMerge(newSense,
				       s, 
				       newSense.similarityTo(s));
		candidateMerges.insert(cand, cand.similarity);
	    }
	} // end while over enumeration of Senses
    }

    /**
     * Removes from candidateMerges all potential merges involving
     * either Sense contained in 'merged'.
     *
     * @param     merged The merged Senses.
     **/
    private void purgeCandidateMerges(CandidateMerge merged) {
	// THIS IS REALLY GROSS AND SHOULD BE REPLACED BY SOMETHING EFFICIENT:
	PriorityQueue pq = new PriorityQueue();
	while (!candidateMerges.isEmpty()) {
	    CandidateMerge cand = (CandidateMerge)candidateMerges.remove();
	    if (cand.s1 != merged.s1 && cand.s1 != merged.s2 &&
		cand.s2 != merged.s1 && cand.s2 != merged.s2) {
		pq.insert(cand, cand.similarity);
	    }
	}
	candidateMerges = pq;
    }

    /**
     * Prior on current word model.
     * <P>
     * Currently we use an exponentially decreasing function of
     * the number of word senses.
     * <P>
     * PERHAPS USE ANDREAS' IDEA TO INCREASE PRIOR WEIGHT DURING
     * TRAINING TO AVOID EARLY OVER-MERGING?
     * <P>
     * Return the base-10 log of this prior.
     *
     * @return    The prior of the Word.
     **/
    private double logPrior() {
	double w = getLexicon().getScenarioSet().getParams().modelPriorWeight();
	double normalization = Math.exp(w) * (1 - Math.exp(-w));
	double prior = normalization * Math.exp(-w * senses.size());
	return Math.log(prior) / LN_TO_LOG;
    }

    /**
     * Likelihood of the set of examples given the word model.
     * Uses log base 10.
     *
     * @return    The log [base 10] of the likelihood of the examples
     *            covered by this Word.
     **/
    private double logLikelihoodOfDataset() {
	double res = 0;
	Enumeration exEnum = examples.elements();
	while (exEnum.hasMoreElements()) {
	    res += Math.log( likelihood( (Fstruct)exEnum.nextElement() ) )
		/ LN_TO_LOG;
	}
	return res;
    }

    /**
     * Sums the likelihoods across all senses, weighted by frequency
     * of each sense.
     * <P>
     * TBD: SHOULD FREQUENCY BE USED ONLY IF USEWORDFREQUENCIES IS
     * ENABLED? 
     * 
     * @param     f The Fstruct to compute likelihood against.
     * @return    The likelihood across all Senses, weighted by Sense
     *            frequency. 
     **/
    private double likelihood(Fstruct f) {
	double res = 0;
	Enumeration sEnum = senses.elements();
	while (sEnum.hasMoreElements()) {
	    Sense s = (Sense)sEnum.nextElement();
	    //          |------------ weight -----------|-> Sense frequency
	    res = res + (s.numExamples() / numExamples()) * s.likelihood(f);
	}
	return res;
    }

    /**
     * Test recognition results ONLY for this word, and report the
     * fraction which are correctly labelled.  The collection of
     * Scenarios which we should recognize against comes from the
     * Dataset's 'testing' set.
     * <P>
     * If there are NO examples to test, a NEGATIVE NUMBER is
     * returned. 
     *
     * @return    The rate of recognition, or a negative number if
     *            there are no examples to recognize against.
     **/
    public double recognitionRate( )
    {
	// THIS CODE IS ADAPTED FROM VERBLEARN.RECOGNIZE()
	int numCorrect = 0;
	int numTested = 0;
	String lang = getLexicon().getLanguage();
	Enumeration scEnum =
	    getLexicon().getScenarioSet().getDataset().testEnum();
	while (scEnum.hasMoreElements()) {
	    Scenario sc = (Scenario)scEnum.nextElement();
	    // ONLY TEST EXAMPLES OF THIS CURRENT WORD
	    if ( sc.isLabeled( lang ) &&
		 sc.getLabel( lang ).equals( getLabel() ) ) {
		VerbComplex actual =
		    getLexicon().label(sc.getUnifiedFstruct());
		numTested++;
		if ( actual.equals(sc.getLabel( lang )) ) {
		    numCorrect++;
		}
	    }
	}
	if (numTested > 0) {
	    return (double)numCorrect / (double)numTested;
	} else {
	    return -1;
	}
    }

    /**
     * Return the closest matching Sense to 'f'.  The 'posterior' field
     * includes a prior reflecting the relative frequency of the sense
     * relative to the 'divisor' parameter, which should be the Slot's
     * total number of examples (for labelling) or the Word's total
     * number of examples (for obeying).  Also note: the label field of
     * the return value is unused.
     * <P>
     * <STRONG>THIS ROUTINE AND ITS RETURN TYPE ARE GETTING TOO
     * OVERLOADED.</STRONG> 
     * 
     * @param     f  The Fstruct to find the closest match to.
     * @param     divisor Context-specific divisor [see comments].
     * @return    The Word Sense which most closely matches f.
     **/
    public CandidateSense bestSenseFor(Fstruct f, double divisor) {
	CandidateSense leading = new CandidateSense();
	leading.posterior = -1;
	Enumeration sEnum = senses.elements();
	while (sEnum.hasMoreElements()) {
	    Sense s = (Sense)sEnum.nextElement();
	    double posterior;
	    if ( getLexicon().getScenarioSet().getParams().
		 useWordFrequencies() ) {
		posterior = (s.numExamples() / divisor) * s.likelihood(f);
	    } else {
		posterior = s.likelihood(f);
	    }
	    if (posterior > leading.posterior) {
		leading.label = getLabel( );
		leading.sense = s;
		leading.posterior = posterior;
	    }
	}
	return leading;
    }

    /**
     * Return posteriors of all Word Senses as a
     * Vector<CandidateSense>. 'divisor' should be the Slot's total
     * number of examples.
     *
     * @param     f The Fstruct to Recognize against.
     * @param     divisor The total number of examples to divide by.
     * @return    A Vector<CandidateSense> which contain the posteriors
     *            given the Fstruct f over all Senses in the Word.
     **/
    public Vector allSensePosteriors(Fstruct f, double divisor) {
	Vector res = new Vector();  // Vector<CandidateSense>
	Enumeration sEnum = getSenses();
	while (sEnum.hasMoreElements()) {
	    Sense s = (Sense)sEnum.nextElement();
	    double pr;
	    if ( getLexicon().getScenarioSet().getParams().
		 useWordFrequencies() ) {
		pr = (s.numExamples() / divisor);
	    } else {
		pr = 1;
	    }
	    double li = s.likelihood( f );
	    res.addElement(new CandidateSense(null, s, pr, li, pr * li));
	}
	return res;
    }

    /**
     * Returns an Enumartions of the Senses in this Word.
     *
     * @return     An Enumeration of the Senses in this Word.
     **/
    public Enumeration getSenses() { return senses.elements(); }

    /**
     * Returns the number of Senses which this Word contains.
     *
     * @return    The number of Senses in this Word.
     **/
    public int senseCount() { return senses.size(); }

    // NC note: this is function to obviate by using a browser
    public String toString() {
	StringBuffer res = new StringBuffer();
	res.append("Model for ").append(getName()).append(" (").append(senses.size())
	    .append(" sense");
	if (senses.size() > 1) res.append("s");
	res.append("):\n");
	Enumeration sEnum = getSenses();
	while (sEnum.hasMoreElements()) {
	    res.append("\n").append(sEnum.nextElement());
	}
	return res.toString();
    }

}


